/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

/* This file describe the implementation in MPI of the communications
   between prpcessors */

#include <strings.h>
#include "pm2v.h"
#include "mpeg2enc.h"


/* This function broadcast the paramater to the slaves processors */
void 
pm2v_broadcastparam (pm2v_stream * stream, char *param)
{
  int length;

  length = strlen (param) + 1;
  MPI_Bcast (&length, 1, MPI_INT, stream->my_id, stream->com);
  MPI_Barrier (stream->com);
  MPI_Bcast (param, length - 1, MPI_BYTE, stream->my_id, stream->com);
}

void 
pm2v_getparam (pm2v_stream * stream, char **param)
{
  int length;
  MPI_Bcast (&length, 1, MPI_INT, stream->master_id, stream->com);

  MPI_Barrier (stream->com);
  *param = (char *) malloc (length * sizeof (char));
  MPI_Bcast (*param, length - 1, MPI_BYTE, stream->master_id, stream->com);
}


void 
pm2v_barrier (pm2v_stream * stream)
{
  MPI_Barrier (stream->com);
}

/* This function broadcast the new repartition of load */
void 
pm2v_bcast_load (pm2v_stream * stream)
{
  MPI_Bcast (stream->nb_macroblock, stream->proc_number, MPI_INT
	     ,stream->master_id, stream->com);
  MPI_Bcast (stream->offset_macroblock, stream->proc_number + 1, MPI_INT
	     ,stream->master_id, stream->com);
  MPI_Bcast (stream->macroblock_to_process, stream->nb_vmacroblock, MPI_INT
	     ,stream->master_id, stream->com);
  MPI_Bcast (stream->idx_macroblock, stream->nb_vmacroblock, MPI_INT
	     ,stream->master_id, stream->com);
}

/* This function gather the performance of each processor */
void 
pm2v_gather_time (pm2v_stream * stream)
{
  MPI_Gather (stream->rate_mcomp + stream->my_id * 10,
	      10,
	      MPI_DOUBLE,
	      stream->rate_mcomp,
	      10,
	      MPI_DOUBLE,
	      stream->master_id,
	      stream->com);
}

/* This function broadcast an image in RGB format 
   (no longer used) */
void
pm2v_broadcast_image (pm2v_stream * stream, char *src)
{
  MPI_Bcast (src, stream->param.vertical_size * stream->param.horizontal_size * 3 * sizeof (char)
	     ,MPI_BYTE, stream->master_id, stream->com);
}


/* This function collect info on each macroblock on the master */
void
pm2v_reduce_macroblock (pm2v_stream * stream)
{
  int i, j, k;

  k = 0;
  // Copy the information for each macroblock in a contiguous block for sending it 
  for (j = 0; j < stream->nb_vmacroblock; j++)
    {
      stream->param.mbinfo[j].fuck = j + 1;
      if (stream->my_id == stream->macroblock_to_process[j])
	{
	  memcpy (stream->buffcopy + k * sizeof (struct mbinfo)
		  ,stream->param.mbinfo + j
		  ,sizeof (struct mbinfo));
	  k++;
	}
    }

  // Gather the informations.
  MPI_Gatherv (stream->buffcopy,
	       stream->nb_macroblock[stream->my_id],
	       stream->type_macroblock,
	       stream->buffcopy2,
	       stream->nb_macroblock,
	       stream->offset_macroblock,
	       stream->type_macroblock,
	       stream->master_id,
	       stream->com);

  
  if (stream->my_id == stream->master_id)
    for (j = 0; j < stream->nb_vmacroblock; j++)
      {
	if (stream->my_id != stream->macroblock_to_process[stream->idx_macroblock[j]])
	  {
	    memcpy (stream->param.mbinfo + stream->idx_macroblock[j]
		    ,stream->buffcopy2 + j * sizeof (struct mbinfo),
		    sizeof (struct mbinfo));
	  }
      }
}


pm2v_reduce_block (pm2v_stream * stream)
{
  int i, j, k;

  k = 0;
  for (j = 0; j < stream->nb_vmacroblock; j++)
    {
      stream->param.mbinfo[j].fuck = j + 1;
      if (stream->my_id == stream->macroblock_to_process[j])
	{
	  memcpy (stream->buffcopy + k * sizeof (short) * 64 * stream->param.block_count
		  , (char *) stream->param.blocks + j * sizeof (short) * 64 * stream->param.block_count
		  ,sizeof (short) * 64 * stream->param.block_count);
	  k++;
	}
    }
  MPI_Gatherv (stream->buffcopy,
	       stream->nb_macroblock[stream->my_id],
	       stream->type_block,
	       stream->buffcopy2,
	       stream->nb_macroblock,
	       stream->offset_macroblock,
	       stream->type_block,
	       stream->master_id,
	       stream->com);
  if (stream->my_id == stream->master_id)
    for (j = 0; j < stream->nb_vmacroblock; j++)
      {
	if (stream->my_id != stream->macroblock_to_process[stream->idx_macroblock[j]])
	  {
	    memcpy ((char *) stream->param.blocks + stream->idx_macroblock[j] * sizeof (short) * 64 * stream->param.block_count
		    ,stream->buffcopy2 + j * sizeof (short) * 64 * stream->param.block_count,
		    sizeof (short) * 64 * stream->param.block_count);
	  }
      }
}


// Collect a plane of an "distributed" image on all processors.
void 
pm2v_reduce_chroma (pm2v_stream * stream, char *src, int chroma)
{
  int i, j, k;

  int mb_chromwidth;
  int mb_chromheight;



  k = 0;
  if (chroma == CHROMA444)
    {
      mb_chromheight = 16;
      mb_chromwidth = 16;
    }
  else
    {
      if (chroma == CHROMA422)
	{
	  mb_chromheight = 16;
	  mb_chromwidth = 8;
	}
      else
	{
	  mb_chromheight = 8;
	  mb_chromwidth = 8;
	}
    }
  for (j = 0; j < stream->nb_vmacroblock; j++)
    {
      stream->param.mbinfo[j].fuck = j + 1;
      if (stream->my_id == stream->macroblock_to_process[j])
	{
	  int x, y;
	  x = j % stream->param.mb_width;
	  y = j / stream->param.mb_width;
	  for (i = 0; i < mb_chromheight; i++)
	    {
	      memcpy (stream->buffcopy + k * mb_chromwidth * mb_chromheight
		      + i * mb_chromheight,
		      src + y * mb_chromheight * mb_chromwidth * stream->param.mb_width
		      + x * mb_chromwidth + i * mb_chromwidth * stream->param.mb_width,
		      mb_chromwidth);
	    }
	  k++;
	}
    }
  if (chroma == CHROMA444)
    MPI_Allgatherv (stream->buffcopy,
		    stream->nb_macroblock[stream->my_id],
		    stream->type_lumi_block,
		    stream->buffcopy2,
		    stream->nb_macroblock,
		    stream->offset_macroblock,
		    stream->type_lumi_block,
		    stream->com);
  else
    {
      if (chroma == CHROMA422)
	MPI_Allgatherv (stream->buffcopy,
			stream->nb_macroblock[stream->my_id],
			stream->type_chroma_block422,
			stream->buffcopy2,
			stream->nb_macroblock,
			stream->offset_macroblock,
			stream->type_chroma_block422,
			stream->com);
      else
	MPI_Allgatherv (stream->buffcopy,
			stream->nb_macroblock[stream->my_id],
			stream->type_chroma_block420,
			stream->buffcopy2,
			stream->nb_macroblock,
			stream->offset_macroblock,
			stream->type_chroma_block420,
			stream->com);
    }
  for (j = 0; j < stream->nb_vmacroblock; j++)
    {
      if (stream->my_id != stream->macroblock_to_process[stream->idx_macroblock[j]])
	{
	  int x, y;
	  x = stream->idx_macroblock[j] % stream->param.mb_width;
	  y = stream->idx_macroblock[j] / stream->param.mb_width;
	  for (i = 0; i < mb_chromheight; i++)
	    memcpy (src + y * mb_chromheight * mb_chromwidth * stream->param.mb_width
	    + x * mb_chromwidth + i * mb_chromwidth * stream->param.mb_width,
		    stream->buffcopy2 + j * mb_chromwidth * mb_chromheight
		    + i * mb_chromheight,
		    mb_chromwidth);
	}
    }
}

// Collect luminance plane on each processors.
void 
pm2v_reduce_lumi (pm2v_stream * stream, char *src)
{
  int i, j, k;

  k = 0;

  for (j = 0; j < stream->nb_vmacroblock; j++)
    {
      stream->param.mbinfo[j].fuck = j + 1;
      if (stream->my_id == stream->macroblock_to_process[j])
	{
	  int x, y;
	  x = j % stream->param.mb_width;
	  y = j / stream->param.mb_width;
	  for (i = 0; i < 16; i++)
	    {
	      memcpy (stream->buffcopy + k * 16 * 16 + i * 16,
		      src + y * 16 * 16 * stream->param.mb_width
		      + x * 16 + i * 16 * stream->param.mb_width,
		      16);
	    }
	  k++;
	}
    }
  MPI_Allgatherv (stream->buffcopy,
		  stream->nb_macroblock[stream->my_id],
		  stream->type_lumi_block,
		  stream->buffcopy2,
		  stream->nb_macroblock,
		  stream->offset_macroblock,
		  stream->type_lumi_block,
		  stream->com);
  for (j = 0; j < stream->nb_vmacroblock; j++)
    {
      if (stream->my_id != stream->macroblock_to_process[stream->idx_macroblock[j]])
	{
	  int x, y;
	  x = stream->idx_macroblock[j] % stream->param.mb_width;
	  y = stream->idx_macroblock[j] / stream->param.mb_width;
	  for (i = 0; i < 16; i++)
	    memcpy (src + y * 16 * 16 * stream->param.mb_width
		    + x * 16 + i * stream->param.mb_width * 16,
		    stream->buffcopy2 + j * 16 * 16 + i * 16,
		    16);
	}
    }
}


// Scatter image on every processors.
void 
pm2v_scatter_image (pm2v_stream * stream, char *src, int width, int height)
{
  int *nb_lines;
  int *offset_lines;
  int i;

  nb_lines = (int *) malloc (stream->proc_number * sizeof (int));
  offset_lines = (int *) malloc ((stream->proc_number + 1) * sizeof (int));

  for (i = 0; i < (stream->proc_number); i++)
    nb_lines[i] = height / stream->proc_number;
  for (i = 0; i < height % stream->proc_number; i++)
    stream->nb_lines[i]++;
  for (i = 0; i < (stream->proc_number); i++)
    nb_lines[i] *= width;
  offset_lines[0] = 0;
  for (i = 1; i < (stream->proc_number + 1); i++)
    offset_lines[i] = offset_lines[i - 1] + nb_lines[i - 1];

  nb_lines[0] += 13 * width;
  for (i = 1; i < (stream->proc_number); i++)
    {
      offset_lines[i] -= 13 * width;
      if (offset_lines[i] < 0)
	offset_lines[i] = 0;
      nb_lines[i] += 26 * width;
      if ((offset_lines[i] + nb_lines[i]) > width * height)
	nb_lines[i] = width * height - offset_lines[i];
    }

  MPI_Scatterv (src,
		nb_lines,
		offset_lines,
		MPI_BYTE,
		src + offset_lines[stream->my_id],
		nb_lines[stream->my_id],
		MPI_BYTE,
		stream->master_id,
		stream->com);

  free (nb_lines);
  free (offset_lines);
}

// Collect and RGB image from each processors to every processors.
void 
pm2v_gather_image (pm2v_stream * stream, char *src, int width, int height)
{
  int *nb_lines;
  int *offset_lines;
  int i;

  nb_lines = (int *) malloc (stream->proc_number * sizeof (int));
  offset_lines = (int *) malloc ((stream->proc_number + 1) * sizeof (int));

  for (i = 0; i < (stream->proc_number); i++)
    nb_lines[i] = height / stream->proc_number;
  for (i = 0; i < height % stream->proc_number; i++)
    nb_lines[i]++;
  for (i = 0; i < (stream->proc_number); i++)
    nb_lines[i] *= width;
  offset_lines[0] = 0;
  for (i = 1; i < (stream->proc_number + 1); i++)
    offset_lines[i] = offset_lines[i - 1] + nb_lines[i - 1];
  MPI_Allgatherv (src + offset_lines[stream->my_id],
		  nb_lines[stream->my_id],
		  MPI_BYTE,
		  src,
		  nb_lines,
		  offset_lines,
		  MPI_BYTE,
		  stream->com);

  free (nb_lines);
  free (offset_lines);
}


void
pm2v_init_com (pm2v_stream * stream)
{
 
  MPI_Type_contiguous (sizeof (struct mbinfo), MPI_BYTE, &(stream->type_macroblock));
  MPI_Type_commit (&(stream->type_macroblock));
  MPI_Type_contiguous (sizeof (short) * 64 * stream->param.block_count, MPI_BYTE, &(stream->type_block));
  MPI_Type_commit (&(stream->type_block));

  MPI_Type_contiguous (16 * 16, MPI_BYTE, &(stream->type_lumi_block));
  MPI_Type_commit (&(stream->type_lumi_block));

  MPI_Type_contiguous (8 * 16, MPI_BYTE, &(stream->type_chroma_block422));
  MPI_Type_commit (&(stream->type_chroma_block422));

  MPI_Type_contiguous (8 * 8, MPI_BYTE, &(stream->type_chroma_block420));
  MPI_Type_commit (&(stream->type_chroma_block420));

}
