/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "quant.h"
#include "conform.h"








/* check for (level independent) parameter limits */
int 
range_checks (pm2v_stream * stream)
{
  int i;

  /* range and value checks */

  if (stream->param.horizontal_size < 1 || stream->param.horizontal_size > 16383)
    {
      stream->local_error = HOR_SIZE_RANGE;
      return -1;
    }
  if (stream->param.mpeg1 && stream->param.horizontal_size > 4095)
    {
      stream->local_error = HOR_SIZE_RANGE;
      return -1;
    }
  if ((stream->param.horizontal_size & 4095) == 0)
    {
      stream->local_error = HOR_SIZE_MOD_4096;
      return -1;
    }
  if (stream->param.chroma_format != CHROMA444 && stream->param.horizontal_size % 2 != 0)
    {
      stream->local_error = HOR_SIZE_NOT_MOD_2;
      return -1;
    }
  if (stream->param.vertical_size < 1 || stream->param.vertical_size > 16383)
    {
      stream->local_error = VERT_SIZE_RANGE;
      return -1;
    }
  if (stream->param.mpeg1 && stream->param.vertical_size > 4095)
    {
      stream->local_error = VERT_SIZE_RANGE;
      return -1;
    }
  if ((stream->param.vertical_size & 4095) == 0)
    {
      stream->local_error = VERT_SIZE_MOD_4096;
      return -1;
    }
  if (stream->param.chroma_format == CHROMA420 && stream->param.vertical_size % 2 != 0)
    {
      stream->local_error = VERT_SIZE_NOT_MOD_2;
      return -1;
    }
  if (stream->param.fieldpic)
    {
      if (stream->param.vertical_size % 2 != 0)
	{
	  stream->local_error = VERT_SIZE_NOT_MOD_2;
	  return -1;
	}
      if (stream->param.chroma_format == CHROMA420 && stream->param.vertical_size % 4 != 0)
	{
	  stream->local_error = VERT_SIZE_NOT_MOD_4;
	  return -1;
	}
    }

  if (stream->param.mpeg1)
    {
      if (stream->param.aspectratio < 1 || stream->param.aspectratio > 14)
	{
	  stream->local_error = VERT_SIZE_NOT_MOD_4;
	  return -1;
	}
    }
  else
    {
      if (stream->param.aspectratio < 1 || stream->param.aspectratio > 4)
	{
	  stream->local_error = BAD_ASPECT_RATIO;
	  return -1;
	}
    }

  if (stream->param.frame_rate_code < 1 || stream->param.frame_rate_code > 8)
    {
      stream->local_error = BAD_FRAME_RATE;
      return -1;
    }
  if (stream->param.bit_rate <= 0.0)
    {
      stream->local_error = BITRATE_RANGE;
      return -1;
    }
  if (stream->param.bit_rate > ((1 << 30) - 1) * 400.0)
    {
      stream->local_error = BITRATE_RANGE;
      return -1;
    }
  if (stream->param.mpeg1 && stream->param.bit_rate > ((1 << 18) - 1) * 400.0)
    {
      stream->local_error = BITRATE_RANGE;
      return -1;
    }
  if (stream->param.vbv_buffer_size < 1 || stream->param.vbv_buffer_size > 0x3ffff)
    {
      stream->local_error = VBV_BUFFER_SIZE_RANGE;
      return -1;
    }
  if (stream->param.mpeg1 && stream->param.vbv_buffer_size >= 1024)
    {
      stream->local_error = VBV_BUFFER_SIZE_RANGE;
      return -1;
    }
  if (stream->param.chroma_format < CHROMA420 || stream->param.chroma_format > CHROMA444)
    {
      stream->local_error = BAD_CHROMA_FORMAT;
      return -1;
    }
  if (stream->param.video_format < 0 || stream->param.video_format > 4)
    {
      stream->local_error = BAD_VIDEO_FORMAT;
      return -1;
    }

  if (stream->param.color_primaries < 1 || stream->param.color_primaries > 7
      || stream->param.color_primaries == 3)
    {
      stream->local_error = BAD_COLOR_PRIMARIES;
      return -1;
    }


  if (stream->param.transfer_characteristics < 1 || stream->param.transfer_characteristics > 7
      || stream->param.transfer_characteristics == 3)
    {
      stream->local_error = TRANSFER_CHARA_RANGE;
      return -1;
    }


  if (stream->param.matrix_coefficients < 1 || stream->param.matrix_coefficients > 7
      || stream->param.matrix_coefficients == 3)
    {
      stream->local_error = MATRIX_COEFF;
      return -1;
    }

  if (stream->param.display_horizontal_size < 0
      || stream->param.display_horizontal_size > 16383)
    {
      stream->local_error = OUTPUT_HOR_SIZE_RANGE;
      return -1;
    }
  if (stream->param.display_vertical_size < 0
      || stream->param.display_vertical_size > 16383)
    {
      stream->local_error = OUTPUT_VERT_SIZE_RANGE;
      return -1;
    }

  if (stream->param.dc_prec < 0 || stream->param.dc_prec > 3)
    {
      stream->local_error = INTRA_DC_PREC_RANGE;
      return -1;
    }

  for (i = 0; i < stream->param.M; i++)
    {
      fprintf (stderr, "%d+%d+%d+%d\n", stream->param.motion_data[i].forw_hor_f_code, stream->param.motion_data[i].forw_hor_f_code, stream->param.motion_data[i].sxf, stream->param.motion_data[i].syf);
      if (stream->param.motion_data[i].forw_hor_f_code < 1
	  || stream->param.motion_data[i].forw_hor_f_code > 9)
	{
	  stream->local_error = F_CODE_RANGE;
	  return -1;
	}
      if (stream->param.motion_data[i].forw_vert_f_code < 1
	  || stream->param.motion_data[i].forw_vert_f_code > 9)
	{
	  stream->local_error = F_CODE_RANGE;
	  return -1;
	}
      if (stream->param.mpeg1 && stream->param.motion_data[i].forw_hor_f_code > 7)
	{
	  stream->local_error = F_CODE_RANGE;
	  return -1;
	}

      if (stream->param.mpeg1 && stream->param.motion_data[i].forw_vert_f_code > 7)
	{
	  stream->local_error = F_CODE_RANGE;
	  return -1;
	}

      if (stream->param.motion_data[i].sxf <= 0)
	{
	  stream->local_error = NEGATIVE_SEARCH_WIN;
	  return -1;
	}
      if (stream->param.motion_data[i].syf <= 0)
	{
	  stream->local_error = NEGATIVE_SEARCH_WIN;
	  return -1;
	}
      if (i != 0)
	{
	  if (stream->param.motion_data[i].back_hor_f_code < 1
	      || stream->param.motion_data[i].back_hor_f_code > 9)
	    {
	      stream->local_error = F_CODE_RANGE;
	      return -1;
	    }
	  if (stream->param.motion_data[i].back_vert_f_code < 1
	      || stream->param.motion_data[i].back_vert_f_code > 9)
	    {
	      stream->local_error = F_CODE_RANGE;
	      return -1;
	    }
	  if (stream->param.mpeg1
	      && stream->param.motion_data[i].back_hor_f_code > 7)
	    {
	      stream->local_error = F_CODE_RANGE;
	      return -1;
	    }
	  if (stream->param.mpeg1
	      && stream->param.motion_data[i].back_vert_f_code > 7)
	    {
	      stream->local_error = F_CODE_RANGE;
	      return -1;
	    }

	  if (stream->param.motion_data[i].sxb <= 0)
	    {
	      stream->local_error = NEGATIVE_SEARCH_WIN;
	      return -1;
	    }
	  if (stream->param.motion_data[i].syb <= 0)
	    {
	      stream->local_error = NEGATIVE_SEARCH_WIN;
	      return -1;
	    }
	}
    }
  return 0;
}

/* identifies valid profile / level combinations */
static char profile_level_defined[5][4] =
{
/* HL   H-14 ML   LL  */
  {1, 1, 1, 0},			/* HP   */
  {0, 1, 0, 0},			/* Spat */
  {0, 0, 1, 1},			/* SNR  */
  {1, 1, 1, 1},			/* MP   */
  {0, 0, 1, 0}			/* SP   */
};

static struct level_limits
  {
    int hor_f_code;
    int vert_f_code;
    int hor_size;
    int vert_size;
    int sample_rate;
    int bit_rate;		/* Mbit/s */
    int vbv_buffer_size;	/* 16384 bit steps */
  }
maxval_tab[4] =
{
  {
    9, 5, 1920, 1152, 62668800, 80, 597
  }
  ,				/* HL */
  {
    9, 5, 1440, 1152, 47001600, 60, 448
  }
  ,				/* H-14 */
  {
    8, 5, 720, 576, 10368000, 15, 112
  }
  ,				/* ML */
  {
    7, 4, 352, 288, 3041280, 4, 29
  }				/* LL */
};

#define SP   5
#define MP   4
#define SNR  3
#define SPAT 2
#define HP   1

#define LL  10
#define ML   8
#define H14  6
#define HL   4

int 
profile_and_level_checks (pm2v_stream * stream)
{
  int i;
  struct level_limits *maxval;

  if (stream->param.profile < 0 || stream->param.profile > 15)
    {
      stream->local_error = BAD_PROFILE;
      return -1;
    }


  if (stream->param.level < 0 || stream->param.level > 15)
    {
      stream->local_error = BAD_LEVEL;
      return -1;
    }


  if (stream->param.profile >= 8)
    {
      //if (!quiet)
      //  fprintf(stderr,"Warning: profile uses a reserved value, conformance checks skipped\n");
      return 0;
    }

  if (stream->param.profile < HP || stream->param.profile > SP)
    {
      stream->local_error = BAD_PROFILE;
      return -1;
    }

  if (stream->param.profile == SNR || stream->param.profile == SPAT)
    {
      stream->local_error = UNSUPPORTED;
      return -1;
    }

  if (stream->param.level < HL || stream->param.level > LL || stream->param.level & 1)
    {
      stream->local_error = BAD_LEVEL;
      return -1;
    }

  maxval = &maxval_tab[(stream->param.level - 4) >> 1];

  /* check profile@level combination */
  if (!profile_level_defined[stream->param.profile - 1][(stream->param.level - 4) >> 1])
    {
      stream->local_error = BAD_PROFILE_LEVEL;
      return -1;
    }


  /* profile (syntax) constraints */

  if (stream->param.profile == SP && stream->param.M != 1)
    {
      stream->local_error = B_PICTURE_FORBIDDEN;
      return -1;
    }

  if (stream->param.profile != HP && stream->param.chroma_format != CHROMA420)
    {
      stream->local_error = BAD_CHROMA_FORMAT_PROFILE;
      return -1;
    }


  if (stream->param.profile == HP && stream->param.chroma_format == CHROMA444)
    {
      stream->local_error = BAD_CHROMA_FORMAT_PROFILE;
      return -1;
    }

  if (stream->param.profile >= MP)	/* SP, MP: constrained repeat_first_field */
    {
      if (stream->param.frame_rate_code <= 2 && stream->param.repeatfirst)
	{
	  stream->local_error = NOT_REPEAT_FIRST_ZERO;;
	  return -1;
	}

      if (stream->param.frame_rate_code <= 6
	  && stream->param.prog_seq && stream->param.repeatfirst)
	{
	  stream->local_error = NOT_REPEAT_FIRST_ZERO;
	  return -1;
	}
    }

  if (stream->param.profile != HP && stream->param.dc_prec == 3)
    {
      stream->local_error = BAD_DC_PRECISION;
      return -1;
    }



  /* level (parameter value) constraints */

  /* Table 8-8 */
  if (stream->param.frame_rate_code > 5 && stream->param.level >= ML)
    {
      stream->local_error = BAD_FRAME_RATE;
      return -1;
    }


  for (i = 0; i < stream->param.M; i++)
    {
      if (stream->param.motion_data[i].forw_hor_f_code > maxval->hor_f_code)
	{
	  stream->local_error = F_CODE_RANGE;
	  return -1;
	}


      if (stream->param.motion_data[i].forw_vert_f_code > maxval->vert_f_code)
	{
	  stream->local_error = F_CODE_RANGE;
	  return -1;
	}

      if (i != 0)
	{
	  if (stream->param.motion_data[i].back_hor_f_code > maxval->hor_f_code)
	    {
	      stream->local_error = F_CODE_RANGE;
	      return -1;
	    }

	  if (stream->param.motion_data[i].back_vert_f_code > maxval->vert_f_code)
	    {
	      stream->local_error = F_CODE_RANGE;
	      return -1;
	    }
	}
    }

  /* Table 8-10 */
  if (stream->param.horizontal_size > maxval->hor_size)
    {
      stream->local_error = HOR_SIZE_RANGE;
      return -1;
    }

  if (stream->param.vertical_size > maxval->vert_size)
    {
      stream->local_error = VERT_SIZE_RANGE;
      return -1;
    }

  /* Table 8-11 */
  if (stream->param.horizontal_size * stream->param.vertical_size * stream->param.frame_rate > maxval->sample_rate)
    {
      stream->local_error = BAD_FRAME_RATE;
      return -1;
    }

  /* Table 8-12 */
  if (stream->param.bit_rate > 1.0e6 * maxval->bit_rate)
    {
      stream->local_error = BITRATE_RANGE;
      return -1;
    }

  /* Table 8-13 */
  if (stream->param.vbv_buffer_size > maxval->vbv_buffer_size)
    {
      stream->local_error = VBV_BUFFER_SIZE_RANGE;
      return -1;
    }
  return 0;
}
