/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include "pm2v.h"


/*****************************************************************************
 pm2v_load_balance

 This function aims to balance the numbers of macroblocs to process along 
 processor. 

 The master has to process less macroblocs due to the overload of putpict.
******************************************************************************/
void 
pm2v_load_balance (pm2v_stream * stream)
{
  double average_time;
  int i, j, k;
  int test;
  int oldmb;
  double temp = 0;
  int min, max;
  double *average_rate;
  double average_putpic;


  // average rate for each processor
  average_rate = (double *) malloc (stream->proc_number * sizeof (double));

  // gather performance for each processor.
  pm2v_gather_time (stream);

  // The master computes the new distribution.
  if (stream->my_id == stream->master_id)
    {

      for (i = 0; i < stream->proc_number; i++)
	{
	  average_rate[i] = 0;
	  for (j = 0; j < 10; j++)
	    average_rate[i] += stream->rate_mcomp[i * 10 + j];
	  average_rate[i] /= 10;
	}


      average_putpic = 0;
      for (j = 0; j < 10; j++)
	average_putpic += stream->time_putpic[j];
      average_putpic /= 10;

#ifdef DEBUG
      fprintf (stderr, "temps pour putpic = %d\n", average_putpic);
#endif

      average_time = 0;
      oldmb = stream->nb_macroblock[stream->master_id];


      for (i = 0; i < stream->proc_number; i++)
	temp += (double) average_rate[i];

      average_time = (stream->nb_vmacroblock
		 + average_putpic * average_rate[stream->master_id]) / temp;


      if (average_time < average_putpic)
	{
	  average_time = (stream->nb_vmacroblock - 1) / (temp - average_rate[0]);
	  stream->nb_macroblock[0] = 1;
	}
      else
	stream->nb_macroblock[0]
	  = (average_time - average_putpic) * average_rate[0];

      for (i = 1; i < stream->proc_number; i++)
	{
	  stream->nb_macroblock[i] = average_time * average_rate[i];
	}


      // Correction of error due to rounding.
      test = 0;
      for (i = 0; i < stream->proc_number; i++)
	{

	  test += stream->nb_macroblock[i];
	  if (stream->nb_macroblock[i] < 1)
	    exit (1);
	}

      test = stream->nb_vmacroblock - test;
      if (test < 0)
	{
	  test = -test;
	  for (i = 0; test > 0; i++)
	    {

	      stream->nb_macroblock[i % stream->proc_number]--;
	      test--;
	      if (stream->nb_macroblock[i % stream->proc_number] < 1)
		{
		  stream->nb_macroblock[i % stream->proc_number] = 1;
		  test++;
		}
	    }
	  test = 0;
	}
      if (test > 0)
	{
	  for (i = 0; i < stream->proc_number; i++)
	    stream->nb_macroblock[i] += test / stream->proc_number;
	  for (i = 0; i < test % stream->proc_number; i++)
	    stream->nb_macroblock[stream->proc_number - 1 - i] += 1;
	}


      // Update information about the new distribution.
      stream->offset_macroblock[0] = 0;
      for (i = 0; i < stream->proc_number; i++)
	{
	  stream->offset_macroblock[i + 1] = stream->nb_macroblock[i] + stream->offset_macroblock[i];
	  for (j = stream->offset_macroblock[i]; j < stream->offset_macroblock[i + 1]; j++)
	    stream->macroblock_to_process[j] = i;
	}


      for (i = 0; i < stream->proc_number; i++)
#ifdef DEBUG
	fprintf (stderr, "nb_macroblock[%d] = %d \n", i, stream->nb_macroblock[i]);
#endif
      k = 0;
      for (j = 0; j < (stream->proc_number); j++)
	for (i = 0; i < stream->nb_vmacroblock; i++)
	  {
	    if (j == stream->macroblock_to_process[i])
	      {
		stream->idx_macroblock[k] = i;
		k++;
	      }
	  }

    }
  // Broadcast the new distribution of macroblocs.
  pm2v_bcast_load (stream);

  free (average_rate);

}
