/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include "pm2v.h"

int
main (int argc, char **argv)
{
  pm2v_stream stream;
  FILE *output;
  int i;
  unsigned char *buffer;	//optimisation pour la lecture des ppm

  char name[256];
  int frame0, nb_of_frame;
  int my_tid;
  char processor_name[MPI_MAX_PROCESSOR_NAME];
  int namelen;
  int done = 0, n, myid, numprocs;

  MPI_Init (&argc, &argv);
  MPI_Comm_size (MPI_COMM_WORLD, &numprocs);
  MPI_Comm_rank (MPI_COMM_WORLD, &myid);
  MPI_Get_processor_name (processor_name, &namelen);

  fprintf (stdout, "Process %d on %s\n",
	   myid, processor_name);


  if (argc != 4)
    {
      MPI_Finalize ();
      printf ("man %s\n", argv[0]);
      exit (1);
    }

  frame0 = atoi (argv[2]);
  nb_of_frame = atoi (argv[3]);


  if (myid != 0)
    output = fopen ("/dev/null", "wb");
  else
    {
      output = fopen ("output.m2v", "wb");
      if (output == NULL)
	{
	  printf ("impossible d'ouvrir le fichier de sortie");
	  exit (1);
	}
    }

  pm2v_open (&stream, "Horizontal_source_size=720;Vertical_source_size=352; Number_of_frames=94;Horizontal_output_size=720;Vertical_output_size=352;ISOIEC_11172_2_stream=0;Interlaced_pictures=0;Bit_rate=10000000;Frame_rate=2;Aspect_ratio=3;Video_format=1;",
	     output, MPI_COMM_WORLD, 0, myid, numprocs);
  buffer = (unsigned char *) malloc (stream.param.height * stream.param.width * 3 * sizeof (unsigned char));
  memset (buffer, 0, stream.param.height * stream.param.width * 3 * sizeof (unsigned char));
  for (i = frame0; i <= nb_of_frame; i++)
    {
      if (myid == 0)
	{
	  sprintf (name, "%s%04i.ppm", argv[1], i);
	  pm2v_readppm (&stream, name, buffer);
	  fprintf (stderr, "Image %i \n", i);
	}
      pm2v_add_image (&stream, buffer);

    }
  pm2v_close (&stream);

  MPI_Finalize ();
  return 0;
}
