/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define GLOBAL			/* used by quant.h */
#include "config.h"
#include "quant.h"
#include "pm2v.h"
#include <assert.h>

#include  "fdctref.h"
#include  "idct.h"
//#include  "motion.h" 
//#include  "mpeg2enc.h" 
//#include  "params.h" 
//#include  "predict.h" 
#include  "putbits.h"
//#include  "puthdr.h" 
//#include  "putmpg.h" 
//#include  "putpic.h" 
//#include  "putseq.h" 
//#include  "putvlc.h" 
//#include  "quantize.h" 
//#include  "ratectl.h" 
//#include  "readpic.h" 
//#include  "stats.h" 
//#include  "transfrm.h"
//#include  "writepic.h" 


/* private prototypes */
void init (pm2v_stream * stream);
void readquantmat (pm2v_stream * stream);




void 
init (pm2v_stream * stream)
{
  int i, j, k, size;
  static int block_count_tab[3] =
  {6, 8, 12};
  int size_chroma_macroblock;

  initbits ();
  init_fdct ();
  init_idct ();

  /* round picture dimensions to nearest multiple of 16 or 32 */
  stream->param.mb_width = (stream->param.horizontal_size + 15) / 16;
  stream->param.mb_height = stream->param.prog_seq ? (stream->param.vertical_size + 15) / 16 : 2 * ((stream->param.vertical_size + 31) / 32);
  stream->param.mb_height2 = stream->param.fieldpic ? stream->param.mb_height >> 1 : stream->param.mb_height;	/* for field pictures */
  stream->param.width = 16 * stream->param.mb_width;
  stream->param.height = 16 * stream->param.mb_height;

  stream->param.chrom_width = (stream->param.chroma_format == CHROMA444) ? stream->param.width : stream->param.width >> 1;
  stream->param.chrom_height = (stream->param.chroma_format != CHROMA420) ? stream->param.height : stream->param.height >> 1;

  stream->param.height2 = stream->param.fieldpic ? stream->param.height >> 1 : stream->param.height;
  stream->param.width2 = stream->param.fieldpic ? stream->param.width << 1 : stream->param.width;
  stream->param.chrom_width2 = stream->param.fieldpic ? stream->param.chrom_width << 1 : stream->param.chrom_width;

  stream->param.block_count = block_count_tab[stream->param.chroma_format - 1];

  /* clip table */
  assert ((stream->param.clp = (unsigned char *) malloc (1024)));

  stream->param.clp += 384;
  for (i = -384; i < 640; i++)
    stream->param.clp[i] = (i < 0) ? 0 : ((i > 255) ? 255 : i);

  for (i = 0; i < 3; i++)
    {
      size = (i == 0) ? stream->param.width * stream->param.height : stream->param.chrom_width * stream->param.chrom_height;

      assert ((stream->param.newrefframe[i] = (unsigned char *) malloc (size)));

      assert ((stream->param.oldrefframe[i] = (unsigned char *) malloc (size)));

      assert ((stream->param.auxframe[i] = (unsigned char *) malloc (size)));

      assert ((stream->param.neworgframe[i] = (unsigned char *) malloc (size)));

      assert ((stream->param.oldorgframe[i] = (unsigned char *) malloc (size)));

      assert ((stream->param.auxorgframe[i] = (unsigned char *) malloc (size)));

      assert ((stream->param.predframe[i] = (unsigned char *) malloc (size)));

    }

  if (stream->param.chroma_format == CHROMA444)
    size_chroma_macroblock = 16 * 16;
  else
    {
      if (stream->param.chroma_format == CHROMA422)
	size_chroma_macroblock = 16 * 8;
      else
	size_chroma_macroblock = 8 * 8;
    }

  stream->nb_vmacroblock = stream->param.mb_width * stream->param.mb_height2;
  // +(stream->proc_number-(stream->param.mb_width*stream->param.mb_height)
  //    %stream->proc_number)%stream->proc_number;
  stream->param.mbinfo = (struct mbinfo *) malloc (stream->nb_vmacroblock * sizeof (struct mbinfo));
  stream->nb_macroblock = (int *) malloc ((stream->proc_number + 1) * sizeof (int));
  stream->nb_lines = (int *) malloc (stream->proc_number * sizeof (int));
  stream->offset_lines = (int *) malloc ((stream->proc_number + 1) * sizeof (int));

  stream->offset_macroblock = (int *) malloc ((stream->proc_number + 1) * sizeof (int));

  /* Buffer of Macroblock to process */
  stream->macroblock_to_process = (int *) malloc ((stream->param.mb_width * stream->param.mb_height2) * sizeof (int));
  stream->idx_macroblock = (int *) malloc ((stream->param.mb_width * stream->param.mb_height2) * sizeof (int));
  stream->offset_macroblock = (int *) malloc ((stream->proc_number + 1) * sizeof (int));
  stream->buffcopy = (char *) malloc (stream->param.mb_width * stream->param.mb_height2 * 16 * 16 * 3);
  stream->buffcopy2 = (char *) malloc (stream->param.mb_width * stream->param.mb_height2 * 16 * 16 * 3);
  stream->rate_mcomp = (double *) malloc (stream->proc_number * 10 * sizeof (double));
  stream->time_putpic = (double *) malloc (10 * sizeof (double));



  /* Interleave macroblocks on each processor to improve load balancing */

  for (i = 0; i < (stream->proc_number); i++)
    {
      stream->nb_macroblock[i] = 0;

    }
  for (i = 0; i < (stream->param.mb_width * stream->param.mb_height2); i++)
    {
      stream->macroblock_to_process[i] = i % (stream->proc_number);
      stream->nb_macroblock[i % (stream->proc_number)]++;

    }
  k = 0;
  stream->offset_macroblock[0] = 0;
  for (j = 1; j <= (stream->proc_number); j++)
    stream->offset_macroblock[j] = stream->offset_macroblock[j - 1] + stream->nb_macroblock[j - 1];
  for (j = 0; j < (stream->proc_number); j++)
    for (i = 0; i < stream->nb_vmacroblock; i++)
      {
	if (j == stream->macroblock_to_process[i])
	  {
	    stream->idx_macroblock[k] = i;
	    k++;
	  }
      }



  for (i = 0; i < (stream->proc_number); i++)
    stream->nb_lines[i] = stream->param.height / stream->proc_number;

  for (i = 0; i < stream->param.height % stream->proc_number; i++)
    stream->nb_lines[i]++;
  stream->offset_lines[0] = 0;


  for (i = 1; i < (stream->proc_number + 1); i++)
    stream->offset_lines[i] = stream->offset_lines[i - 1] + stream->nb_lines[i - 1];

  assert (stream->param.mbinfo);


  stream->param.blocks =
    (short (*)[64]) malloc (stream->param.mb_width * stream->param.mb_height2 * stream->param.block_count * sizeof (short[64]));

  assert (stream->param.blocks);



}



void 
readquantmat (pm2v_stream * stream)
{
  int i;
  //int v;
  //FILE *fd;

  //if (iqname[0]=='-')
  {
    /* use default intra matrix */
    stream->param.load_iquant = 0;
    for (i = 0; i < 64; i++)
      stream->param.intra_q[i] = default_intra_quantizer_matrix[i];
  }
  /*
     else
     {
     // read customized intra matrix 
     load_iquant = 1;
     if (!(fd = fopen(iqname,"r")))
     {
     sprintf(errortext,"Couldn't open quant matrix file %s",iqname);
     error(errortext);
     }

     for (i=0; i<64; i++)
     {
     fscanf(fd,"%d",&v);
     if (v<1 || v>255)
     error("invalid value in quant matrix");
     intra_q[i] = v;
     }

     fclose(fd);
     }
   */

//if (niqname[0]=='-')
  {
    // use default non-intra matrix 
    stream->param.load_niquant = 0;
    for (i = 0; i < 64; i++)
      stream->param.inter_q[i] = 16;
  }
/*
   else
   {
   // read customized non-intra matrix 
   load_niquant = 1;
   if (!(fd = fopen(niqname,"r")))
   {
   sprintf(errortext,"Couldn't open quant matrix file %s",niqname);
   error(errortext);
   }

   for (i=0; i<64; i++)
   {
   fscanf(fd,"%d",&v);
   if (v<1 || v>255)
   error("invalid value in quant matrix");
   inter_q[i] = v;


   fclose(fd);
   }
 */
}
