/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/


/* 
This is the description of the parser for analysing strings of arguments 
supplied to pm2v_open 
*/

%{
#define ANALYSE
#include <stdlib.h>
#include <strings.h>
#include <stdio.h>
#include <assert.h>
#include "config.h"
#include "pm2v.h"
#include "conform.h"
#include "mpeg2enc.h"
#include "params.h"
#include "params_init.h"

static char * stream_addr;

int yyerror(char *);
int yylex();

%}


%union {
type_expression exp;
char * variable_name;
char * string;
}


%token <string> CONSTANT
%token <string> STRING
%token <string> VARIABLE

%token BOOLEAN INTEGER FLOAT
%token EOF

%right '='
%left '+' '-' 
%left '*' '/'
%right '(' ')'
%right ','

%type <exp> expression
%type <exp> list_expression 
%type <exp> oned_expression
%type <exp> list_oned_expression
%type <exp> twod_expression

%start config

%%

config: 
| list_declaration 
;

list_declaration:
declaration
| declaration list_declaration 
;


declaration:
VARIABLE '=' STRING ';' {
  int i;
  int found=0;
  for (i=0;i<(sizeof(paramParse)/sizeof(paramParse[0]));i++) {
    if (!strcmp(paramParse[i].name,$1)) {
      found=1;
      
      if (paramParse[i].type==STRING) {
	*((char**)(stream_addr+paramParse[i].variable-(int)&dummy_stream))=malloc(strlen($3)+1);
	$3[strlen($3)-1]=0;
 	strcpy(*((char**)(stream_addr+paramParse[i].variable-(int)&dummy_stream)),$3+1);
      }
    }
    if (!found)
      yyerror("Symbol not found\n");
  }
}
| VARIABLE '=' expression ';' {
  int i;
  int found=0;
  for (i=0;i<(sizeof(paramParse)/sizeof(paramParse[0]));i++) {
    if (!strcmp(paramParse[i].name,$1)) {
      found=1;
      if (paramParse[i].type==INTEGER)
	*(int*)(stream_addr+paramParse[i].variable-(int)&dummy_stream)=(int)*$3.value;
      else
	if (paramParse[i].type==FLOAT)
	  *(double*)(stream_addr+paramParse[i].variable-(int)&dummy_stream)=(double)*$3.value;
	else
	  yyerror("");
    }
  }
  if (!found)
    yyerror("Symbol not found\n");
}
| VARIABLE '=' oned_expression ';' 
{
  int i;
  int found=0;
  for (i=0;i<(sizeof(paramParse)/sizeof(paramParse[0]));i++) {
    if (!strcmp(paramParse[i].name,$1)) {
      found=1;
      if (paramParse[i].type==INTEGER)
	{
	  int j;
	  *((int**)(stream_addr+paramParse[i].variable-(int)&dummy_stream))=(int*)malloc($3.size_alloc*sizeof(int));
	  for (j=0;j<$3.size_alloc;j++) {
	    *((int*)(stream_addr+paramParse[i].variable-(int)&dummy_stream)+j)=(int)(*($3.value+j));
	}
    }
      else
	if (paramParse[i].type==FLOAT)
	  {
	    int j;
	    *((double**)(stream_addr+paramParse[i].variable-(int)&dummy_stream))=(double*)malloc($3.size_alloc*sizeof(double));
	    for (j=0;j<$3.size_alloc;j++)
	      *((double*)(stream_addr+paramParse[i].variable-(int)&dummy_stream)+j)=*(double*)($3.value+j);
	  }
	else
	  yyerror("");
    }
  }
  if (!found)
    yyerror("Symbol not found\n");
}
| VARIABLE '=' twod_expression ';' 
{
  int i;
  int found=0;
  for (i=0;i<(sizeof(paramParse)/sizeof(paramParse[0]));i++) {
    if (!strcmp(paramParse[i].name,$1)) {
      found=1;
      if (paramParse[i].type==INTEGER)
	{
	  // Affect the variable rerenced in paramParse
	  int j;
	  *((int**)(stream_addr+paramParse[i].variable-(int)&dummy_stream))=(int*)malloc($3.size_alloc*sizeof(int));
	  for (j=0;j<$3.size_alloc;j++) {
	    *(*(int**)(stream_addr+paramParse[i].variable-(int)&dummy_stream)+j)=((int)(*($3.value+j)));
	    }
	}
      else
	if (paramParse[i].type==FLOAT)
	  {
	    int j;
	    *((double**)(stream_addr+paramParse[i].variable-(int)&dummy_stream))=(double*)malloc($3.size_alloc*sizeof(double));
	    for (j=0;i<$3.size_alloc;j++)
	      *(*(double**)(stream_addr+paramParse[i].variable-(int)&dummy_stream)+j)=*(double*)($3.value+j);
	  }
	else
	  yyerror("");
    }
  }
  if (!found)
    yyerror("Symbol not found\n");
}
;

/*
lvalue:
VARIABLE 
  | VARIABLE '[' INTEGER ']' | VARIABLE '[' INTEGER ',' INTEGER ']' 
;
*/

twod_expression:
'{' list_oned_expression '}' {
  int i,j;
  for (i=0;i<$2.dim2;i++){
    for (j=0;j<$2.dim1;j++){
      //fprintf(stderr,"%g ",$2.value[i*$2.dim1+j]); 
    }
    fprintf(stderr,"\n");
  }
  $$=$2;
} 
;

list_oned_expression:
list_oned_expression ',' oned_expression {
  $$.dim2=++$1.dim2;
  $$.value=realloc($1.value,($1.size_alloc+$1.dim1)*sizeof(double));
  memcpy($$.value+$$.size_alloc,$3.value,$3.dim1*sizeof(double));  
  $$.size_alloc=$1.size_alloc+$1.dim1;
} 
| oned_expression { $$=$1; }
;


oned_expression:
'{' list_expression '}' {
  $$=$2;
}
;

list_expression:
list_expression ',' expression {
  $$.dim1=++$1.dim1;
  $$.value=realloc($1.value,($1.size_alloc+1)*sizeof(double));
  $$.value[$$.size_alloc]=*$3.value;
  $$.size_alloc=$1.size_alloc+1;
}
| expression {$$=$1;}
;


expression:
CONSTANT {
                    $$.value=(double*)malloc(sizeof(double));
		    *$$.value=atof($1);
		    $$.dim1=1;
		    $$.dim2=1;
		    $$.size_alloc=1;
		    }
| expression '+' expression {*$$.value = *$1.value + *$3.value;}
| expression '*' expression {*$$.value = *$1.value * *$3.value;}
| expression '-' expression {*$$.value = *$1.value - *$3.value;}
| expression '/' expression {*$$.value = *$1.value / *$3.value;}
| '(' expression ')' {*$$.value = *$2.value;}
;

%%

char * string_to_parse;
int string_pointer;


#define YY_INPUT(buf,result,max_size) \
   if (string_pointer!=-1) { \
   if ((strlen(string_to_parse+string_pointer)) <= max_size) { \
      strcpy(buf,string_to_parse+string_pointer); \
      result=strlen(string_to_parse+string_pointer); \
      string_pointer=-1; }\
   else { \
      strncpy(buf,string_to_parse+string_pointer,max_size); \
      result=max_size; \
      string_pointer+=max_size; \
      }\
   } \
   else { result=0; }

   

#include "lex.yy.c"



int parseparam(char * string)
{
string_to_parse=(char*)malloc(strlen(string)+1);
strcpy(string_to_parse,string);
string_pointer=0;
return !yyparse();
}



int loadparam(pm2v_stream * stream,char * param)
{
  int i;
  static double ratetab[8]=
    {24000.0/1001.0,24.0,25.0,30000.0/1001.0,30.0,50.0,60000.0/1001.0,60.0};

  stream_addr=(char*)stream;
  assert(parseparam(default_param));
  if (!parseparam(param))
    return -1;

  if (stream->param.N<1) {
    stream->local_error=BAD_N;
    return -1; }
  if (stream->param.M<1) {
    stream->local_error=BAD_N;
    return -1; }
  if (stream->param.N%stream->param.M != 0)  {
    stream->local_error=BAD_N;
    return -1; }
 
  /* make flags boolean (x!=0 -> x=1) */
  stream->param.mpeg1 = !!stream->param.mpeg1;
  stream->param.fieldpic = !!stream->param.fieldpic;
  stream->param.low_delay = !!stream->param.low_delay;
  stream->param.constrparms = !!stream->param.constrparms;
  stream->param.prog_seq = !!stream->param.prog_seq;
  stream->param.topfirst = !!stream->param.topfirst;

  for (i=0; i<3; i++)
  {
    stream->param.frame_pred_dct_tab[i] = !!stream->param.frame_pred_dct_tab[i];
    stream->param.conceal_tab[i] = !!stream->param.conceal_tab[i];
    stream->param.qscale_tab[i] = !!stream->param.qscale_tab[i];
    stream->param.intravlc_tab[i] = !!stream->param.intravlc_tab[i];
    stream->param.altscan_tab[i] = !!stream->param.altscan_tab[i];
  }
  stream->param.repeatfirst = !!stream->param.repeatfirst;
  stream->param.prog_frame = !!stream->param.prog_frame;

  /* make sure MPEG specific parameters are valid */
  if (range_checks(stream)) {
    fprintf(stderr,"Error %d\n",stream->local_error);
  }

  stream->param.frame_rate = ratetab[stream->param.frame_rate_code-1];

  /* timecode -> frame number */

  if (!stream->param.mpeg1)
  {
    if (fprintf(stderr,"-------> %i\n",profile_and_level_checks(stream)));
  }
  else
  {
    /* MPEG-1 */
    if (stream->param.constrparms)
    {
      if (stream->param.horizontal_size>768
          || stream->param.vertical_size>576
          || ((stream->param.horizontal_size+15)/16)*((stream->param.vertical_size+15)/16)>396
          || ((stream->param.horizontal_size+15)/16)*((stream->param.vertical_size+15)/16)*stream->param.frame_rate>396*25.0
          || stream->param.frame_rate>30.0)
      {
        stream->param.constrparms = 0;
      }
    }

    if (stream->param.constrparms)
    {
      for (i=0; i<stream->param.M; i++)
      {
        if (stream->param.motion_data[i].forw_hor_f_code>4)
        {
          stream->param.constrparms = 0;
          break;
        }

        if (stream->param.motion_data[i].forw_vert_f_code>4)
        {
          stream->param.constrparms = 0;
          break;
        }

        if (i!=0)
        {
          if (stream->param.motion_data[i].back_hor_f_code>4)
          {
            stream->param.constrparms = 0;
            break;
          }

          if (stream->param.motion_data[i].back_vert_f_code>4)
          {
            stream->param.constrparms = 0;
            break;
          }
        }
      }
    }
  }

  /* relational checks */

  if (stream->param.mpeg1)
  {
    if (!stream->param.prog_seq)
    {
      stream->param.prog_seq = 1;
    }

    if (stream->param.chroma_format!=CHROMA420)
    {
      stream->param.chroma_format = CHROMA420;
    }

    if (stream->param.dc_prec!=0)
    {
        stream->param.dc_prec = 0;
    }

    for (i=0; i<3; i++)
      if (stream->param.qscale_tab[i])
      {
	stream->param.qscale_tab[i] = 0;
      }

    for (i=0; i<3; i++)
      if (stream->param.intravlc_tab[i])
      {
                stream->param.intravlc_tab[i] = 0;
      }

    for (i=0; i<3; i++)
      if (stream->param.altscan_tab[i])
      {
               stream->param.altscan_tab[i] = 0;
      }
  }

  if (!stream->param.mpeg1 && stream->param.constrparms)
  {
        stream->param.constrparms = 0;
  }

  if (stream->param.prog_seq && !stream->param.prog_frame)
  {
        stream->param.prog_frame = 1;
  }

  if (stream->param.prog_frame && stream->param.fieldpic)
  {
       stream->param.fieldpic = 0;
  }

  if (!stream->param.prog_frame && stream->param.repeatfirst)
  {
    stream->param.repeatfirst = 0;
  }

  if (stream->param.prog_frame)
  {
    for (i=0; i<3; i++)
      if (!stream->param.frame_pred_dct_tab[i])
      {
        stream->param.frame_pred_dct_tab[i] = 1;
      }
  }

  if (stream->param.prog_seq && !stream->param.repeatfirst && stream->param.topfirst)
  {
    stream->param.topfirst = 0;
  }

  /* search windows */
  for (i=0; i<stream->param.M; i++)
  {
    if (stream->param.motion_data[i].sxf > (4<<stream->param.motion_data[i].forw_hor_f_code)-1)
    {
      stream->param.motion_data[i].sxf = (4<<stream->param.motion_data[i].forw_hor_f_code)-1;
    }

    if (stream->param.motion_data[i].syf > (4<<stream->param.motion_data[i].forw_vert_f_code)-1)
    {
      stream->param.motion_data[i].syf = (4<<stream->param.motion_data[i].forw_vert_f_code)-1;
    }

    if (i!=0)
    {
      if (stream->param.motion_data[i].sxb > (4<<stream->param.motion_data[i].back_hor_f_code)-1)
      {
        stream->param.motion_data[i].sxb = (4<<stream->param.motion_data[i].back_hor_f_code)-1;
      }

      if (stream->param.motion_data[i].syb > (4<<stream->param.motion_data[i].back_vert_f_code)-1)
      {
        stream->param.motion_data[i].syb = (4<<stream->param.motion_data[i].back_vert_f_code)-1;
      }
    }
  }
  return 0;
}









