/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

static pm2v_stream dummy_stream; 

/* Structure describing the keywords used by the string of configuration.*/

struct {
  char * name; // Name of the keyword.
  int type;    // Type of the variable (integer,float,string)
  int variable; // Offset of the variable in pm2v_stream.
  int dim1;  // not used yet.
  int dim2;  // not used yet.
} paramParse[]= 
{
  {"Copyright_info",STRING,(int)&dummy_stream.param.id_string,1,1},
  {"Number_of_frames",INTEGER,(int)&dummy_stream.param.nframes,1,1},
  {"Number_of_first_frame",INTEGER,(int)&dummy_stream.param.frame0,1,1},
  {"Timecode_of_first_frame",INTEGER,(int)&dummy_stream.param.tc0,1,1},
  {"Number_of_frame_in_GOP",INTEGER,(int)&dummy_stream.param.N,1,1},
  {"IP_frame_distance",INTEGER,(int)&dummy_stream.param.M,1,1},
  {"ISOIEC_11172_2_stream",INTEGER,(int)&dummy_stream.param.mpeg1,1,1},
  {"Interlaced_pictures",INTEGER,(int)&dummy_stream.param.fieldpic,1,1},
  {"Horizontal_source_size",INTEGER,(int)&dummy_stream.param.horizontal_size,1,1},
  {"Vertical_source_size",INTEGER,(int)&dummy_stream.param.vertical_size,1,1},
  {"Aspect_ratio",INTEGER,(int)&dummy_stream.param.aspectratio,1,1},
  {"Frame_rate",INTEGER,(int)&dummy_stream.param.frame_rate_code,1,1},
  {"Bit_rate",FLOAT,(int)&dummy_stream.param.bit_rate,1,1},
  {"Vbv_buffer_size",INTEGER,(int)&dummy_stream.param.vbv_buffer_size,1,1},
  {"Low_delay",INTEGER,(int)&dummy_stream.param.low_delay,1,1},
  {"Constrained_parameter_flags",INTEGER,(int)&dummy_stream.param.constrparms,1,1},
  {"Profile",INTEGER,(int)&dummy_stream.param.profile,1,1},
  {"Level_ID",INTEGER,(int)&dummy_stream.param.level,1,1},
  {"Progressive_sequence",INTEGER,(int)&dummy_stream.param.prog_seq,1,1},
  {"Chroma_format",INTEGER,(int)&dummy_stream.param.chroma_format,1,1},
  {"Video_format",INTEGER,(int)&dummy_stream.param.video_format,1,1},
  {"Color_primaries",INTEGER,(int)&dummy_stream.param.color_primaries,1,1},
  {"Transfer_charasteristics",INTEGER,(int)&dummy_stream.param.transfer_characteristics,1,1},
  {"Matrix_coefficients",INTEGER,(int)&dummy_stream.param.matrix_coefficients,1,1},
  {"Horizontal_output_size",INTEGER,(int)&dummy_stream.param.display_horizontal_size,1,1},
  {"Vertical_output_size",INTEGER,(int)&dummy_stream.param.display_vertical_size,1,1},
  {"Intra_DC_precision",INTEGER,(int)&dummy_stream.param.dc_prec,1,1},
  {"Top_field_first",INTEGER,(int)&dummy_stream.param.topfirst,1,1},
  {"Frame_prediction_frame_dct",INTEGER,(int)&dummy_stream.param.frame_pred_dct_tab,3,1},
  {"Concealment_motion_vectors",INTEGER,(int)&dummy_stream.param.conceal_tab,3,1},
  {"Quantization_scale_type",INTEGER,(int)&dummy_stream.param.qscale_tab,3,1},
  {"Intra_VLC_format",INTEGER,(int)&dummy_stream.param.intravlc_tab,3,1},
  {"Alternate_scan",INTEGER,(int)&dummy_stream.param.altscan_tab,3,1},
  {"Repeat_first_field",INTEGER,(int)&dummy_stream.param.repeatfirst,1,1},
  {"Progressive_frame",INTEGER,(int)&dummy_stream.param.prog_frame,1,1},
  {"P_distance",INTEGER,(int)&dummy_stream.param.P,1,1},
  {"Rate_control_r",INTEGER,(int)&dummy_stream.param.r,1,1},
  {"Rate_control_avg_act",INTEGER,(int)&dummy_stream.param.avg_act,1,1},
  {"Rate_control_Xi",INTEGER,(int)&dummy_stream.param.Xi,1,1},
  {"Rate_control_Xp",INTEGER,(int)&dummy_stream.param.Xp,1,1},
  {"Rate_control_Xb",INTEGER,(int)&dummy_stream.param.Xb,1,1},
  {"Rate_control_d0i",INTEGER,(int)&dummy_stream.param.d0i,1,1},
  {"Rate_control_d0p",INTEGER,(int)&dummy_stream.param.d0p,1,1},
  {"Rate_control_d0b",INTEGER,(int)&dummy_stream.param.d0b,1,1},
  {"Motion_vectors_parameters",INTEGER,(int)&dummy_stream.param.motion_data,4,0}
};

// String describing the defaults paramaters of the encoder.

 static char * default_param=""
   "Copyright_info=\"P2MV, Parallel MPEG2 encoder based on mpeg2encode\";"
   "Number_of_frames=1000;"
   "Number_of_first_frame=0;"
   "Timecode_of_first_frame=0;"
   "Number_of_frame_in_GOP=12;"
   "IP_frame_distance=3;"
   "ISOIEC_11172_2_stream=0;"
   "Interlaced_pictures=0;"
   "Horizontal_source_size=704;"
   "Vertical_source_size=576;"
   "Aspect_ratio=2;"
   "Frame_rate=3;"
   "Bit_rate=10000000;"
   "Vbv_buffer_size=112;"
   "Low_delay=0;"
   "Constrained_parameter_flags=0;"
   "Profile=4;"
   "Level_ID=8;"
   "Progressive_sequence=1;"
   "Chroma_format=1;"
   "Video_format=1;"
   "Color_primaries=5;"
   "Transfer_charasteristics=5;"
   "Matrix_coefficients=5;"
   "Horizontal_output_size=704;"
   "Vertical_output_size=576;"
   "Intra_DC_precision=0;"
   "Top_field_first=0;"
   "Frame_prediction_frame_dct={0,0,0};"
   "Concealment_motion_vectors={0,0,0};"
   "Quantization_scale_type={1,1,1};"
   "Intra_VLC_format={1,0,0};"
   "Alternate_scan={0,0,0};"
   "Repeat_first_field=0;"
   "Progressive_frame=0;"
   "P_distance=0;"
   "Rate_control_r=0;"
   "Rate_control_avg_act=0;"
   "Rate_control_Xi=0;"
   "Rate_control_Xp=0;"
   "Rate_control_Xb=0;"
   "Rate_control_d0i=0;"
   "Rate_control_d0p=0;"
   "Rate_control_d0b=0;"  
   "Motion_vectors_parameters={{02,02,11,11},"
                              "{00,00,00,00},"
                              "{01,01,03,03},"
                              "{01,01,07,07},"
                              "{01,01,07,07},"
                              "{01,01,03,03}};";





