/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <mpi.h>
#include "config.h"
#include "quant.h"
#include "pm2v.h"

#include  "fdctref.h" 
#include  "idct.h" 
#include  "motion.h" 
#include  "mpeg2enc.h" 
#include  "params.h" 
#include  "predict.h" 
#include  "putbits.h" 
#include  "puthdr.h" 
#include  "putmpg.h" 
#include  "putpic.h" 
#include  "putseq.h" 
#include  "putvlc.h" 
#include  "quantize.h" 
#include  "ratectl.h" 
#include  "readpic.h" 
#include  "stats.h" 
#include  "transfrm.h"
#include  "writepic.h" 

#include "assert.h"


int pm2v_open(pm2v_stream* stream,char * param,FILE * output,pm2v_communicator com,int master_id
	      ,int my_id,int nb_proc) {
  int i;
  /* read parameter file */

  stream->com=com;
  stream->my_id=my_id;
  stream->com=com;
  stream->proc_number=nb_proc;
  stream->master_id=master_id;
  stream->picture_count=0;
  if (master_id==my_id)
    {
      pm2v_broadcastparam(stream,param);
      loadparam(stream,param);
    }
  else
    {
      fprintf(stderr,"teuprou\n");
      pm2v_getparam(stream,&param);
      fprintf(stderr,"prouteu\n");
      loadparam(stream,param);
    }
  fprintf(stderr,"my_id %i\n",stream->my_id);
  stream->param.output=output;
  
  /* read quantization matrices */
  readquantmat(stream);

  init(stream);
  rc_init_seq(stream); /* initialize rate control */

  /* sequence header, sequence extension and sequence display extension */
  putseqhdr(stream);
  if (!stream->param.mpeg1)
    {
      putseqext(stream);
      putseqdispext(stream);
    }
  
  /* optionally output some text data (description, copyright or whatever) */
  if (strlen(stream->param.id_string) > 1)
    putuserdata(stream,stream->param.id_string);
  pm2v_init_com(stream);

  /*initialization of the picture pool*/
  stream->pictures_pool = (unsigned char ***) malloc((stream->param.M)*sizeof(unsigned char*));
  for(i=0;i<stream->param.M;i++) {
    stream->pictures_pool[i]=(unsigned char **) malloc(3*sizeof(unsigned char*)*stream->param.width*stream->param.height);
    assert(stream->pictures_pool[i][0]= 
       (unsigned char *) malloc(stream->param.width*stream->param.height*sizeof(unsigned char))); 
    stream->pictures_pool[i][1]= 
       (unsigned char *) malloc(stream->param.width*stream->param.height*sizeof(unsigned char));
    stream->pictures_pool[i][2]= 
       (unsigned char *) malloc(stream->param.width*stream->param.height*sizeof(unsigned char)); 
  }
  
  return 0;
}


/* close the input stream */
void pm2v_close(pm2v_stream* stream){
  int i;

  for(i=0;i<stream->param.M;i++) {
    free(stream->pictures_pool[i][0]);
    free(stream->pictures_pool[i][1]);
    free(stream->pictures_pool[i][2]); 
    free(stream->pictures_pool[i]);
  }
  free(stream->pictures_pool);

  putseqend(stream);
}

/* read a ppm file */
void pm2v_readppm(pm2v_stream* stream,char* filename, unsigned char* output){
	FILE* f;
	int i;
		
	if(!(f=fopen(filename,"rb"))){
		printf("Can't open %s ! \n",filename);
		exit(1);
	}
	
	getc(f); getc(f); /* magic number (P6) */
	pbm_getint(f); pbm_getint(f); pbm_getint(f); /* width height maxcolors */
	for (i=0;i<stream->param.vertical_size;i++)
	  fread(output+i*stream->param.width2*3,1,stream->param.horizontal_size*3,f);	
	fclose(f);
	
}

/* add an image to the input stream */
/* thatfor we use a buffer of M images called pictures_pool */
void pm2v_add_image(pm2v_stream* stream,unsigned char* img){
	static int i=0,j,k;
	int tmp;
	long long ref;
	struct timeval tp;

        // Convert the image in YUV and add it in picture pool.
	my_readframe(stream,stream->pictures_pool[i%(stream->param.M)],img);
	if(i%(stream->param.M)==0){
	  if(i==0) {
	    putseq(stream);
	  }
	  else 
	    for(j=0;j<stream->param.M;j++) {
	      if (stream->picture_count>10) {
                // If we have enough B Frame encoded we can estimate the inbalance.
		if (stream->param.pict_type==B_TYPE)
		  pm2v_load_balance(stream); 
	      }
	      if (stream->my_id==stream->master_id) {
		 gettimeofday(&tp,NULL);
		 ref=(long long)tp.tv_sec*1000000+tp.tv_usec;
		 putpict(stream,stream->neworg);
		 gettimeofday(&tp,NULL);
		 stream->time_putpic[stream->picture_count%10]
		   =((long long)tp.tv_sec*1000000+tp.tv_usec-ref)/100;
		
	      }
	      putseq(stream);
	      if (stream->param.pict_type==B_TYPE)
		stream->picture_count++;
	    }
	}
	i++;
}































