/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include  "fdctref.h"
#include  "idct.h"
#include  "motion.h"
#include  "mpeg2enc.h"
#include  "params.h"
#include  "predict.h"
#include  "putbits.h"
#include  "puthdr.h"
#include  "putmpg.h"
#include  "putpic.h"
#include  "putseq.h"
#include  "putvlc.h"
#include  "quantize.h"
#include  "ratectl.h"
#include  "readpic.h"
#include  "stats.h"
#include  "transfrm.h"
#include  "writepic.h"
#include <stdio.h>
#include <math.h>
#include "config.h"
#include "quant.h"

/* private prototypes */
static int frametotc (pm2v_stream * stream, int frame);

/* generate sequence header (6.2.2.1, 6.3.3)

 * matrix download not implemented
 */
void 
putseqhdr (pm2v_stream * stream)
{
  int i;

  alignbits (stream);
  putbits (stream, SEQ_START_CODE, 32);		/* sequence_header_code */
  putbits (stream, stream->param.horizontal_size, 12);	/* horizontal_size_value */
  putbits (stream, stream->param.vertical_size, 12);	/* vertical_size_value */
  putbits (stream, stream->param.aspectratio, 4);	/* aspect_ratio_information */
  putbits (stream, stream->param.frame_rate_code, 4);	/* frame_rate_code */
  putbits (stream, (int) ceil (stream->param.bit_rate / 400.0), 18);	/* bit_rate_value */
  putbits (stream, 1, 1);	/* marker_bit */
  putbits (stream, stream->param.vbv_buffer_size, 10);	/* vbv_buffer_size_value */
  putbits (stream, stream->param.constrparms, 1);	/* constrained_parameters_flag */

  putbits (stream, stream->param.load_iquant, 1);	/* load_intra_quantizer_matrix */
  if (stream->param.load_iquant)
    for (i = 0; i < 64; i++)	/* matrices are always downloaded in zig-zag order */
      putbits (stream, stream->param.intra_q[zig_zag_scan[i]], 8);	/* intra_quantizer_matrix */

  putbits (stream, stream->param.load_niquant, 1);	/* load_non_intra_quantizer_matrix */
  if (stream->param.load_niquant)
    for (i = 0; i < 64; i++)
      putbits (stream, stream->param.inter_q[zig_zag_scan[i]], 8);	/* non_intra_quantizer_matrix */
}

/* generate sequence extension (6.2.2.3, 6.3.5) header (MPEG-2 only) */
void 
putseqext (pm2v_stream * stream)
{
  alignbits (stream);
  putbits (stream, EXT_START_CODE, 32);		/* extension_start_code */
  putbits (stream, SEQ_ID, 4);	/* extension_start_code_identifier */
  putbits (stream, (stream->param.profile << 4) | stream->param.level, 8);	/* profile_and_level_indication */
  putbits (stream, stream->param.prog_seq, 1);	/* progressive sequence */
  putbits (stream, stream->param.chroma_format, 2);	/* chroma_format */
  putbits (stream, stream->param.horizontal_size >> 12, 2);	/* horizontal_size_extension */
  putbits (stream, stream->param.vertical_size >> 12, 2);	/* vertical_size_extension */
  putbits (stream, ((int) ceil (stream->param.bit_rate / 400.0)) >> 18, 12);	/* bit_rate_extension */
  putbits (stream, 1, 1);	/* marker_bit */
  putbits (stream, stream->param.vbv_buffer_size >> 10, 8);	/* vbv_buffer_size_extension */
  putbits (stream, 0, 1);	/* low_delay  -- currently not implemented */
  putbits (stream, 0, 2);	/* frame_rate_extension_n */
  putbits (stream, 0, 5);	/* frame_rate_extension_d */
}

/* generate sequence display extension (6.2.2.4, 6.3.6)

 * content not yet user setable
 */
void 
putseqdispext (pm2v_stream * stream)
{
  alignbits (stream);
  putbits (stream, EXT_START_CODE, 32);		/* extension_start_code */
  putbits (stream, DISP_ID, 4);	/* extension_start_code_identifier */
  putbits (stream, stream->param.video_format, 3);	/* video_format */
  putbits (stream, 1, 1);	/* colour_description */
  putbits (stream, stream->param.color_primaries, 8);	/* colour_primaries */
  putbits (stream, stream->param.transfer_characteristics, 8);	/* transfer_characteristics */
  putbits (stream, stream->param.matrix_coefficients, 8);	/* matrix_coefficients */
  putbits (stream, stream->param.display_horizontal_size, 14);	/* display_horizontal_size */
  putbits (stream, 1, 1);	/* marker_bit */
  putbits (stream, stream->param.display_vertical_size, 14);	/* display_vertical_size */
}

/* output a zero terminated string as user data (6.2.2.2.2, 6.3.4.1)

 * string must not emulate start codes
 */
void 
putuserdata (pm2v_stream * stream, char *userdata)
{
  alignbits (stream);
  putbits (stream, USER_START_CODE, 32);	/* user_data_start_code */
  while (*userdata)
    putbits (stream, *userdata++, 8);
}

/* generate group of pictures header (6.2.2.6, 6.3.9)

 * uses tc0 (timecode of first frame) and frame0 (number of first frame)
 */
void 
putgophdr (pm2v_stream * stream, int frame, int closed_gop)
{
  int tc;

  alignbits (stream);
  putbits (stream, GOP_START_CODE, 32);		/* group_start_code */
  tc = frametotc (stream, stream->param.tc0 + frame);
  putbits (stream, tc, 25);	/* time_code */
  putbits (stream, closed_gop, 1);	/* closed_gop */
  putbits (stream, 0, 1);	/* broken_link */
}

/* convert frame number to time_code

 * drop_frame not implemented
 */
static int 
frametotc (pm2v_stream * stream, int frame)
{
  int fps, pict, sec, minute, hour, tc;

  fps = (int) (stream->param.frame_rate + 0.5);
  pict = frame % fps;
  frame = (frame - pict) / fps;
  sec = frame % 60;
  frame = (frame - sec) / 60;
  minute = frame % 60;
  frame = (frame - minute) / 60;
  hour = frame % 24;
  tc = (hour << 19) | (minute << 13) | (1 << 12) | (sec << 6) | pict;

  return tc;
}

/* generate picture header (6.2.3, 6.3.10) */
void 
putpicthdr (pm2v_stream * stream)
{
  alignbits (stream);
  putbits (stream, PICTURE_START_CODE, 32);	/* picture_start_code */
  calc_vbv_delay (stream);
  putbits (stream, stream->param.temp_ref, 10);		/* temporal_reference */
  putbits (stream, stream->param.pict_type, 3);		/* picture_coding_type */
  putbits (stream, stream->param.vbv_delay, 16);	/* vbv_delay */

  if (stream->param.pict_type == P_TYPE || stream->param.pict_type == B_TYPE)
    {
      putbits (stream, 0, 1);	/* full_pel_forward_vector */
      if (stream->param.mpeg1)
	putbits (stream, stream->param.forw_hor_f_code, 3);
      else
	putbits (stream, 7, 3);	/* forward_f_code */
    }

  if (stream->param.pict_type == B_TYPE)
    {
      putbits (stream, 0, 1);	/* full_pel_backward_vector */
      if (stream->param.mpeg1)
	putbits (stream, stream->param.back_hor_f_code, 3);
      else
	putbits (stream, 7, 3);	/* backward_f_code */
    }

  putbits (stream, 0, 1);	/* extra_bit_picture */
}

/* generate picture coding extension (6.2.3.1, 6.3.11)

 * composite display information (v_axis etc.) not implemented
 */
void 
putpictcodext (pm2v_stream * stream)
{
  alignbits (stream);
  putbits (stream, EXT_START_CODE, 32);		/* extension_start_code */
  putbits (stream, CODING_ID, 4);	/* extension_start_code_identifier */
  putbits (stream, stream->param.forw_hor_f_code, 4);	/* forward_horizontal_f_code */
  putbits (stream, stream->param.forw_vert_f_code, 4);	/* forward_vertical_f_code */
  putbits (stream, stream->param.back_hor_f_code, 4);	/* backward_horizontal_f_code */
  putbits (stream, stream->param.back_vert_f_code, 4);	/* backward_vertical_f_code */
  putbits (stream, stream->param.dc_prec, 2);	/* intra_dc_precision */
  putbits (stream, stream->param.pict_struct, 2);	/* picture_structure */
  putbits (stream, (stream->param.pict_struct == FRAME_PICTURE) ? stream->param.topfirst : 0, 1);	/* top_field_first */
  putbits (stream, stream->param.frame_pred_dct, 1);	/* frame_pred_frame_dct */
  putbits (stream, 0, 1);	/* concealment_motion_vectors  -- currently not implemented */
  putbits (stream, stream->param.q_scale_type, 1);	/* q_scale_type */
  putbits (stream, stream->param.intravlc, 1);	/* intra_vlc_format */
  putbits (stream, stream->param.altscan, 1);	/* alternate_scan */
  putbits (stream, stream->param.repeatfirst, 1);	/* repeat_first_field */
  putbits (stream, stream->param.prog_frame, 1);	/* chroma_420_type */
  putbits (stream, stream->param.prog_frame, 1);	/* progressive_frame */
  putbits (stream, 0, 1);	/* composite_display_flag */
}

/* generate sequence_end_code (6.2.2) */
void 
putseqend (pm2v_stream * stream)
{
  alignbits (stream);
  putbits (stream, SEQ_END_CODE, 32);
}
