/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/


#include  "fdctref.h"
#include  "idct.h"
#include  "motion.h"
#include  "mpeg2enc.h"
#include  "params.h"
#include  "predict.h"
#include  "putbits.h"
#include  "puthdr.h"
#include  "putmpg.h"
#include  "putpic.h"
#include  "putseq.h"
#include  "putvlc.h"
#include  "quantize.h"
#include  "ratectl.h"
#include  "readpic.h"
#include  "stats.h"
#include  "transfrm.h"
#include  "writepic.h"
#include <stdio.h>
#include "config.h"
#include "quant.h"

/* generate variable length codes for an intra-coded block (6.2.6, 6.3.17) */
void 
putintrablk (pm2v_stream * stream, short *blk, int cc)
{
  int n, dct_diff, run, signed_level;

  /* DC coefficient (7.2.1) */
  dct_diff = blk[0] - stream->param.dc_dct_pred[cc];	/* difference to previous block */
  stream->param.dc_dct_pred[cc] = blk[0];

  if (cc == 0)
    putDClum (stream, dct_diff);
  else
    putDCchrom (stream, dct_diff);

  /* AC coefficients (7.2.2) */
  run = 0;
  for (n = 1; n < 64; n++)
    {
      /* use appropriate entropy scanning pattern */
      signed_level = blk[(stream->param.altscan ? alternate_scan : zig_zag_scan)[n]];
      if (signed_level != 0)
	{
	  putAC (stream, run, signed_level, stream->param.intravlc);
	  run = 0;
	}
      else
	run++;			/* count zero coefficients */
    }

  /* End of Block -- normative block punctuation */
  if (stream->param.intravlc)
    putbits (stream, 6, 4);	/* 0110 (Table B-15) */
  else
    putbits (stream, 2, 2);	/* 10 (Table B-14) */
}

/* generate variable length codes for a non-intra-coded block (6.2.6, 6.3.17) */
void 
putnonintrablk (pm2v_stream * stream, short *blk)
{
  int n, run, signed_level, first;

  run = 0;
  first = 1;

  for (n = 0; n < 64; n++)
    {
      /* use appropriate entropy scanning pattern */
      signed_level = blk[(stream->param.altscan ? alternate_scan : zig_zag_scan)[n]];

      if (signed_level != 0)
	{
	  if (first)
	    {
	      /* first coefficient in non-intra block */
	      putACfirst (stream, run, signed_level);
	      first = 0;
	    }
	  else
	    putAC (stream, run, signed_level, 0);

	  run = 0;
	}
      else
	run++;			/* count zero coefficients */
    }

  /* End of Block -- normative block punctuation  */
  putbits (stream, 2, 2);
}

/* generate variable length code for a motion vector component (7.6.3.1) */
void 
putmv (pm2v_stream * stream, int dmv, int f_code)
{
  int r_size, f, vmin, vmax, dv, temp, motion_code, motion_residual;

  r_size = f_code - 1;		/* number of fixed length code ('residual') bits */
  f = 1 << r_size;
  vmin = -16 * f;		/* lower range limit */
  vmax = 16 * f - 1;		/* upper range limit */
  dv = 32 * f;

  /* fold vector difference into [vmin...vmax] */
  if (dmv > vmax)
    dmv -= dv;
  else if (dmv < vmin)
    dmv += dv;


  /* split dmv into motion_code and motion_residual */
  temp = ((dmv < 0) ? -dmv : dmv) + f - 1;
  motion_code = temp >> r_size;
  if (dmv < 0)
    motion_code = -motion_code;
  motion_residual = temp & (f - 1);

  putmotioncode (stream, motion_code);	/* variable length code */

  if (r_size != 0 && motion_code != 0)
    putbits (stream, motion_residual, r_size);	/* fixed length code */
}
