/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/


#include  "fdctref.h"
#include  "idct.h"
#include  "motion.h"
#include  "mpeg2enc.h"
#include  "params.h"
#include  "predict.h"
#include  "putbits.h"
#include  "puthdr.h"
#include  "putmpg.h"
#include  "putpic.h"
#include  "putseq.h"
#include  "putvlc.h"
#include  "quantize.h"
#include  "ratectl.h"
#include  "readpic.h"
#include  "stats.h"
#include  "transfrm.h"
#include  "writepic.h"
#include <stdio.h>
#include "config.h"
#include "quant.h"

/* private prototypes */
static void putmvs (pm2v_stream * stream, int MV[2][2][2], int PMV[2][2][2],
	    int mv_field_sel[2][2], int dmvector[2], int s, int motion_type,
		    int hor_f_code, int vert_f_code);

/* quantization / variable length encoding of a complete picture */
void 
putpict (pm2v_stream * stream, unsigned char *frame)
{
  int i, j, k, comp, cc;
  int mb_type;
  int PMV[2][2][2];
  int prev_mquant;
  int cbp, MBAinc = 0;


  //pm2v_reduce_macroblock(stream);
  //pm2v_barrier(stream);
  //if (stream->my_id==stream->master_id)
  {

    rc_init_pict (stream, frame);	/* set up rate control */

    /* picture header and picture coding extension */
    putpicthdr (stream);

    if (!stream->param.mpeg1)
      putpictcodext (stream);

    prev_mquant = rc_start_mb (stream);		/* initialize quantization parameter */

    k = 0;

    for (j = 0; j < stream->param.mb_height2; j++)
      {
	/* macroblock row loop */

	for (i = 0; i < stream->param.mb_width; i++)
	  {
	    /* macroblock loop */
	    if (i == 0)
	      {
		/* slice header (6.2.4) */
		alignbits (stream);

		if (stream->param.mpeg1 || stream->param.vertical_size <= 2800)
		  putbits (stream, SLICE_MIN_START + j, 32);	/* slice_start_code */
		else
		  {
		    putbits (stream, SLICE_MIN_START + (j & 127), 32);	/* slice_start_code */
		    putbits (stream, j >> 7, 3);	/* slice_vertical_position_extension */
		  }

		/* quantiser_scale_code */
		putbits (stream, stream->param.q_scale_type ? map_non_linear_mquant[prev_mquant]
			 : prev_mquant >> 1, 5);

		putbits (stream, 0, 1);		/* extra_bit_slice */

		/* reset predictors */

		for (cc = 0; cc < 3; cc++)
		  stream->param.dc_dct_pred[cc] = 0;

		PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;
		PMV[0][1][0] = PMV[0][1][1] = PMV[1][1][0] = PMV[1][1][1] = 0;

		MBAinc = i + 1;	/* first MBAinc denotes absolute position */
	      }

	    mb_type = stream->param.mbinfo[k].mb_type;

	    /* determine mquant (rate control) */
	    stream->param.mbinfo[k].mquant = rc_calc_mquant (stream, k);

	    /* quantize macroblock */
	    if (mb_type & MB_INTRA)
	      {
		for (comp = 0; comp < stream->param.block_count; comp++)
		  quant_intra (stream, stream->param.blocks[k * stream->param.block_count + comp], stream->param.blocks[k * stream->param.block_count + comp],
			       stream->param.dc_prec, stream->param.intra_q, stream->param.mbinfo[k].mquant);
		stream->param.mbinfo[k].cbp = cbp = (1 << stream->param.block_count) - 1;
	      }
	    else
	      {
		cbp = 0;
		for (comp = 0; comp < stream->param.block_count; comp++)
		  cbp = (cbp << 1) | quant_non_intra (stream, stream->param.blocks[k * stream->param.block_count + comp],
		  stream->param.blocks[k * stream->param.block_count + comp],
		     stream->param.inter_q, stream->param.mbinfo[k].mquant);

		stream->param.mbinfo[k].cbp = cbp;

		if (cbp)
		  mb_type |= MB_PATTERN;
	      }

	    /* output mquant if it has changed */
	    if (cbp && prev_mquant != stream->param.mbinfo[k].mquant)
	      mb_type |= MB_QUANT;

	    /* check if macroblock can be skipped */
	    if (i != 0 && i != stream->param.mb_width - 1 && !cbp)
	      {
		/* no DCT coefficients and neither first nor last macroblock of slice */

		if (stream->param.pict_type == P_TYPE && !(mb_type & MB_FORWARD))
		  {
		    /* P picture, no motion vectors -> skip */

		    /* reset predictors */

		    for (cc = 0; cc < 3; cc++)
		      stream->param.dc_dct_pred[cc] = 0;

		    PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;
		    PMV[0][1][0] = PMV[0][1][1] = PMV[1][1][0] = PMV[1][1][1] = 0;

		    stream->param.mbinfo[k].mb_type = mb_type;
		    stream->param.mbinfo[k].skipped = 1;
		    MBAinc++;
		    k++;
		    continue;
		  }

		if (stream->param.pict_type == B_TYPE && stream->param.pict_struct == FRAME_PICTURE
		    && stream->param.mbinfo[k].motion_type == MC_FRAME
		    && ((stream->param.mbinfo[k - 1].mb_type ^ mb_type) & (MB_FORWARD | MB_BACKWARD)) == 0
		    && (!(mb_type & MB_FORWARD) ||
		     (PMV[0][0][0] == stream->param.mbinfo[k].MV[0][0][0] &&
		      PMV[0][0][1] == stream->param.mbinfo[k].MV[0][0][1]))
		    && (!(mb_type & MB_BACKWARD) ||
		     (PMV[0][1][0] == stream->param.mbinfo[k].MV[0][1][0] &&
		      PMV[0][1][1] == stream->param.mbinfo[k].MV[0][1][1])))
		  {
		    /* conditions for skipping in B frame pictures:
		     * - must be frame predicted
		     * - must be the same prediction type (forward/backward/interp.)
		     *   as previous macroblock
		     * - relevant vectors (forward/backward/both) have to be the same
		     *   as in previous macroblock
		     */

		    stream->param.mbinfo[k].mb_type = mb_type;
		    stream->param.mbinfo[k].skipped = 1;
		    MBAinc++;
		    k++;
		    continue;
		  }

		if (stream->param.pict_type == B_TYPE && stream->param.pict_struct != FRAME_PICTURE
		    && stream->param.mbinfo[k].motion_type == MC_FIELD
		    && ((stream->param.mbinfo[k - 1].mb_type ^ mb_type) & (MB_FORWARD | MB_BACKWARD)) == 0
		    && (!(mb_type & MB_FORWARD) ||
		     (PMV[0][0][0] == stream->param.mbinfo[k].MV[0][0][0] &&
		      PMV[0][0][1] == stream->param.mbinfo[k].MV[0][0][1] &&
		      stream->param.mbinfo[k].mv_field_sel[0][0] == (stream->param.pict_struct == BOTTOM_FIELD)))
		    && (!(mb_type & MB_BACKWARD) ||
		     (PMV[0][1][0] == stream->param.mbinfo[k].MV[0][1][0] &&
		      PMV[0][1][1] == stream->param.mbinfo[k].MV[0][1][1] &&
		      stream->param.mbinfo[k].mv_field_sel[0][1] == (stream->param.pict_struct == BOTTOM_FIELD))))
		  {
		    /* conditions for skipping in B field pictures:
		     * - must be field predicted
		     * - must be the same prediction type (forward/backward/interp.)
		     *   as previous macroblock
		     * - relevant vectors (forward/backward/both) have to be the same
		     *   as in previous macroblock
		     * - relevant motion_vertical_field_selects have to be of same
		     *   parity as current field
		     */

		    stream->param.mbinfo[k].mb_type = mb_type;
		    stream->param.mbinfo[k].skipped = 1;
		    MBAinc++;
		    k++;
		    continue;
		  }
	      }

	    /* macroblock cannot be skipped */
	    stream->param.mbinfo[k].skipped = 0;

	    /* there's no VLC for 'No MC, Not Coded':
	     * we have to transmit (0,0) motion vectors
	     */
	    if (stream->param.pict_type == P_TYPE && !cbp && !(mb_type & MB_FORWARD))
	      mb_type |= MB_FORWARD;

	    putaddrinc (stream, MBAinc);	/* macroblock_address_increment */
	    MBAinc = 1;

	    putmbtype (stream, stream->param.pict_type, mb_type);	/* macroblock type */

	    if (mb_type & (MB_FORWARD | MB_BACKWARD) && !stream->param.frame_pred_dct)
	      putbits (stream, stream->param.mbinfo[k].motion_type, 2);

	    if (stream->param.pict_struct == FRAME_PICTURE && cbp && !stream->param.frame_pred_dct)
	      putbits (stream, stream->param.mbinfo[k].dct_type, 1);

	    if (mb_type & MB_QUANT)
	      {
		putbits (stream, stream->param.q_scale_type ? map_non_linear_mquant[stream->param.mbinfo[k].mquant]
			 : stream->param.mbinfo[k].mquant >> 1, 5);
		prev_mquant = stream->param.mbinfo[k].mquant;
	      }

	    if (mb_type & MB_FORWARD)
	      {
		/* forward motion vectors, update predictors */
		putmvs (stream, stream->param.mbinfo[k].MV, PMV, stream->param.mbinfo[k].mv_field_sel, stream->param.mbinfo[k].dmvector, 0,
			stream->param.mbinfo[k].motion_type, stream->param.forw_hor_f_code, stream->param.forw_vert_f_code);
	      }

	    if (mb_type & MB_BACKWARD)
	      {
		/* backward motion vectors, update predictors */
		putmvs (stream, stream->param.mbinfo[k].MV, PMV, stream->param.mbinfo[k].mv_field_sel, stream->param.mbinfo[k].dmvector, 1,
			stream->param.mbinfo[k].motion_type, stream->param.back_hor_f_code, stream->param.back_vert_f_code);
	      }

	    if (mb_type & MB_PATTERN)
	      {
		putcbp (stream, (cbp >> (stream->param.block_count - 6)) & 63);
		if (stream->param.chroma_format != CHROMA420)
		  putbits (stream, cbp, stream->param.block_count - 6);
	      }

	    for (comp = 0; comp < stream->param.block_count; comp++)
	      {
		/* block loop */
		if (cbp & (1 << (stream->param.block_count - 1 - comp)))
		  {
		    if (mb_type & MB_INTRA)
		      {
			cc = (comp < 4) ? 0 : (comp & 1) + 1;
			putintrablk (stream, stream->param.blocks[k * stream->param.block_count + comp], cc);
		      }
		    else
		      putnonintrablk (stream, stream->param.blocks[k * stream->param.block_count + comp]);
		  }
	      }

	    /* reset predictors */
	    if (!(mb_type & MB_INTRA))
	      for (cc = 0; cc < 3; cc++)
		stream->param.dc_dct_pred[cc] = 0;

	    if (mb_type & MB_INTRA || (stream->param.pict_type == P_TYPE && !(mb_type & MB_FORWARD)))
	      {
		PMV[0][0][0] = PMV[0][0][1] = PMV[1][0][0] = PMV[1][0][1] = 0;
		PMV[0][1][0] = PMV[0][1][1] = PMV[1][1][0] = PMV[1][1][1] = 0;
	      }

	    stream->param.mbinfo[k].mb_type = mb_type;
	    k++;
	  }
      }

    rc_update_pict (stream);
    vbv_end_of_picture ();
  }
}

/* output motion vectors (6.2.5.2, 6.3.16.2)

 * this routine also updates the predictions for motion vectors (PMV)
 */

static void 
putmvs (pm2v_stream * stream, int MV[2][2][2], int PMV[2][2][2], int mv_field_sel[2][2], int dmvector[2], int s, int motion_type, int hor_f_code, int vert_f_code)
{
  if (stream->param.pict_struct == FRAME_PICTURE)
    {
      if (motion_type == MC_FRAME)
	{
	  /* frame prediction */
	  putmv (stream, MV[0][s][0] - PMV[0][s][0], hor_f_code);
	  putmv (stream, MV[0][s][1] - PMV[0][s][1], vert_f_code);
	  PMV[0][s][0] = PMV[1][s][0] = MV[0][s][0];
	  PMV[0][s][1] = PMV[1][s][1] = MV[0][s][1];
	}
      else if (motion_type == MC_FIELD)
	{
	  /* field prediction */
	  putbits (stream, mv_field_sel[0][s], 1);
	  putmv (stream, MV[0][s][0] - PMV[0][s][0], hor_f_code);
	  putmv (stream, (MV[0][s][1] >> 1) - (PMV[0][s][1] >> 1), vert_f_code);
	  putbits (stream, mv_field_sel[1][s], 1);
	  putmv (stream, MV[1][s][0] - PMV[1][s][0], hor_f_code);
	  putmv (stream, (MV[1][s][1] >> 1) - (PMV[1][s][1] >> 1), vert_f_code);
	  PMV[0][s][0] = MV[0][s][0];
	  PMV[0][s][1] = MV[0][s][1];
	  PMV[1][s][0] = MV[1][s][0];
	  PMV[1][s][1] = MV[1][s][1];
	}
      else
	{
	  /* dual prime prediction */
	  putmv (stream, MV[0][s][0] - PMV[0][s][0], hor_f_code);
	  putdmv (stream, dmvector[0]);
	  putmv (stream, (MV[0][s][1] >> 1) - (PMV[0][s][1] >> 1), vert_f_code);
	  putdmv (stream, dmvector[1]);
	  PMV[0][s][0] = PMV[1][s][0] = MV[0][s][0];
	  PMV[0][s][1] = PMV[1][s][1] = MV[0][s][1];
	}
    }
  else
    {
      /* field picture */
      if (motion_type == MC_FIELD)
	{
	  /* field prediction */
	  putbits (stream, mv_field_sel[0][s], 1);
	  putmv (stream, MV[0][s][0] - PMV[0][s][0], hor_f_code);
	  putmv (stream, MV[0][s][1] - PMV[0][s][1], vert_f_code);
	  PMV[0][s][0] = PMV[1][s][0] = MV[0][s][0];
	  PMV[0][s][1] = PMV[1][s][1] = MV[0][s][1];
	}
      else if (motion_type == MC_16X8)
	{
	  /* 16x8 prediction */
	  putbits (stream, mv_field_sel[0][s], 1);
	  putmv (stream, MV[0][s][0] - PMV[0][s][0], hor_f_code);
	  putmv (stream, MV[0][s][1] - PMV[0][s][1], vert_f_code);
	  putbits (stream, mv_field_sel[1][s], 1);
	  putmv (stream, MV[1][s][0] - PMV[1][s][0], hor_f_code);
	  putmv (stream, MV[1][s][1] - PMV[1][s][1], vert_f_code);
	  PMV[0][s][0] = MV[0][s][0];
	  PMV[0][s][1] = MV[0][s][1];
	  PMV[1][s][0] = MV[1][s][0];
	  PMV[1][s][1] = MV[1][s][1];
	}
      else
	{
	  /* dual prime prediction */
	  putmv (stream, MV[0][s][0] - PMV[0][s][0], hor_f_code);
	  putdmv (stream, dmvector[0]);
	  putmv (stream, MV[0][s][1] - PMV[0][s][1], vert_f_code);
	  putdmv (stream, dmvector[1]);
	  PMV[0][s][0] = PMV[1][s][0] = MV[0][s][0];
	  PMV[0][s][1] = PMV[1][s][1] = MV[0][s][1];
	}
    }
}
