/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/


#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include "config.h"
#include "quant.h"
#include "pm2v.h"


#include  "fdctref.h"
#include  "idct.h"
#include  "motion.h"
#include  "mpeg2enc.h"
#include  "params.h"
#include  "predict.h"
#include  "putbits.h"
#include  "puthdr.h"
#include  "putmpg.h"
#include  "putpic.h"
#include  "putseq.h"
#include  "putvlc.h"
#include  "quantize.h"
#include  "ratectl.h"
#include  "readpic.h"
#include  "stats.h"
#include  "transfrm.h"
#include  "writepic.h"


void 
putseq (pm2v_stream * stream)
{
  /* this routine assumes (N % M) == 0 */
  static int i = 0, j, k, f, f0, n, np, nb, sxf, syf, sxb, syb;
  static int ipflag;
  static unsigned char *newref[3], *neworg[3];

  /* f0: lowest frame number in current GOP

   * first GOP contains N-(M-1) frames,
   * all other GOPs contain N frames
   */
  f0 = stream->param.N
    * ((i + (stream->param.M - 1)) / stream->param.N) - (stream->param.M - 1);

  if (f0 < 0)
    f0 = 0;

  if (i == 0 || (i - 1) % stream->param.M == 0)
    {
      /* I or P frame */
      for (j = 0; j < 3; j++)
	{
	  /* shuffle reference frames */
	  neworg[j] = stream->param.oldorgframe[j];
	  newref[j] = stream->param.oldrefframe[j];
	  stream->param.oldorgframe[j] = stream->param.neworgframe[j];
	  stream->param.oldrefframe[j] = stream->param.newrefframe[j];
	  stream->param.neworgframe[j] = neworg[j];
	  stream->param.newrefframe[j] = newref[j];
	}

      /* f: frame number in display order */
      f = (i == 0) ? 0 : i + stream->param.M - 1;
      if (f >= stream->param.nframes)
	f = stream->param.nframes - 1;

      if (i == f0)		/* first displayed frame in GOP is I */
	{
	  /* I frame */
	  stream->param.pict_type = I_TYPE;
	  stream->param.forw_hor_f_code = stream->param.forw_vert_f_code = 15;
	  stream->param.back_hor_f_code = stream->param.back_vert_f_code = 15;

	  /* n: number of frames in current GOP

	   * first GOP contains (M-1) less (B) frames
	   */
	  n = (i == 0) ? stream->param.N - (stream->param.M - 1) : stream->param.N;

	  /* last GOP may contain less frames */
	  if (n > stream->param.nframes - f0)
	    n = stream->param.nframes - f0;

	  /* number of P frames */
	  if (i == 0)
	    np = (n + 2 * (stream->param.M - 1)) / stream->param.M - 1;		/* first GOP */
	  else
	    np = (n + (stream->param.M - 1)) / stream->param.M - 1;

	  /* number of B frames */
	  nb = n - np - 1;

	  rc_init_GOP (stream, np, nb);

	  putgophdr (stream, f0, i == 0);	/* set closed_GOP in first GOP only */
	}
      else
	{
	  /* P frame */
	  stream->param.pict_type = P_TYPE;
	  stream->param.forw_hor_f_code = stream->param.motion_data[0].forw_hor_f_code;
	  stream->param.forw_vert_f_code = stream->param.motion_data[0].forw_vert_f_code;
	  stream->param.back_hor_f_code = stream->param.back_vert_f_code = 15;
	  sxf = stream->param.motion_data[0].sxf;
	  syf = stream->param.motion_data[0].syf;
	}
    }
  else
    {
      /* B frame */
      for (j = 0; j < 3; j++)
	{
	  neworg[j] = stream->param.auxorgframe[j];
	  newref[j] = stream->param.auxframe[j];
	}

      /* f: frame number in display order */
      f = i - 1;
      stream->param.pict_type = B_TYPE;
      n = (i - 2) % stream->param.M + 1;	/* first B: n=1, second B: n=2, ... */
      stream->param.forw_hor_f_code = stream->param.motion_data[n].forw_hor_f_code;
      stream->param.forw_vert_f_code = stream->param.motion_data[n].forw_vert_f_code;
      stream->param.back_hor_f_code = stream->param.motion_data[n].back_hor_f_code;
      stream->param.back_vert_f_code = stream->param.motion_data[n].back_vert_f_code;
      sxf = stream->param.motion_data[n].sxf;
      syf = stream->param.motion_data[n].syf;
      sxb = stream->param.motion_data[n].sxb;
      syb = stream->param.motion_data[n].syb;
    }

  stream->param.temp_ref = f - f0;
  stream->param.frame_pred_dct = stream->param.frame_pred_dct_tab[stream->param.pict_type - 1];
  stream->param.q_scale_type = stream->param.qscale_tab[stream->param.pict_type - 1];
  stream->param.intravlc = stream->param.intravlc_tab[stream->param.pict_type - 1];
  stream->param.altscan = stream->param.altscan_tab[stream->param.pict_type - 1];


  //my_readframe(stream,neworg, stream->pictures_pool[f%(stream->param.M)]);
  //neworg[0]=stream->pictures_pool[f%(stream->param.M)][0];
  //neworg[1]=stream->pictures_pool[f%(stream->param.M)][1];
  //neworg[2]=stream->pictures_pool[f%(stream->param.M)][2];
  memcpy (neworg[0], stream->pictures_pool[f % (stream->param.M)][0],
	  stream->param.width2 * stream->param.height2);

  if (stream->param.chroma_format == CHROMA444)
    {
      memcpy (neworg[1], stream->pictures_pool[f % (stream->param.M)][1], stream->param.width2 * stream->param.height2);
      memcpy (neworg[2], stream->pictures_pool[f % (stream->param.M)][2], stream->param.width2 * stream->param.height2);
    }
  if (stream->param.chroma_format == CHROMA422)
    {
      memcpy (neworg[1], stream->pictures_pool[f % (stream->param.M)][1], stream->param.width2 * stream->param.height2 / 2);
      memcpy (neworg[2], stream->pictures_pool[f % (stream->param.M)][2], stream->param.width2 * stream->param.height2 / 2);
    }
  if (stream->param.chroma_format == CHROMA420)
    {
      memcpy (neworg[1], stream->pictures_pool[f % (stream->param.M)][1], stream->param.width2 * stream->param.height2 / 4);
      memcpy (neworg[2], stream->pictures_pool[f % (stream->param.M)][2], stream->param.width2 * stream->param.height2 / 4);
    }
  if (stream->param.fieldpic)
    {

      stream->param.pict_struct = stream->param.topfirst ? TOP_FIELD : BOTTOM_FIELD;

      motion_estimation (stream, stream->param.oldorgframe[0], stream->param.neworgframe[0],
		 stream->param.oldrefframe[0], stream->param.newrefframe[0],
			 neworg[0], newref[0],
			 sxf, syf, sxb, syb, stream->param.mbinfo, 0, 0);



      predict (stream, stream->param.oldrefframe, stream->param.newrefframe, stream->param.predframe, 0, stream->param.mbinfo);
      dct_type_estimation (stream, stream->param.predframe[0], neworg[0], stream->param.mbinfo);
      transform (stream, stream->param.predframe, neworg, stream->param.mbinfo, stream->param.blocks);

      stream->neworg = neworg[0];
      putpict (stream, neworg[0]);

      for (k = 0; k < stream->param.mb_height2 * stream->param.mb_width; k++)
	{
	  if (stream->param.mbinfo[k].mb_type & MB_INTRA)
	    for (j = 0; j < stream->param.block_count; j++)
	      iquant_intra (stream, stream->param.blocks[k * stream->param.block_count + j], stream->param.blocks[k * stream->param.block_count + j],
			    stream->param.dc_prec, stream->param.intra_q, stream->param.mbinfo[k].mquant);
	  else
	    for (j = 0; j < stream->param.block_count; j++)
	      iquant_non_intra (stream, stream->param.blocks[k * stream->param.block_count + j], stream->param.blocks[k * stream->param.block_count + j],
		     stream->param.inter_q, stream->param.mbinfo[k].mquant);
	}

      itransform (stream, stream->param.predframe, newref, stream->param.mbinfo, stream->param.blocks);
      calcSNR (stream, neworg, newref);
      stats (stream);


      stream->param.pict_struct = stream->param.topfirst ? BOTTOM_FIELD : TOP_FIELD;

      ipflag = (stream->param.pict_type == I_TYPE);
      if (ipflag)
	{
	  /* first field = I, second field = P */
	  stream->param.pict_type = P_TYPE;
	  stream->param.forw_hor_f_code = stream->param.motion_data[0].forw_hor_f_code;
	  stream->param.forw_vert_f_code = stream->param.motion_data[0].forw_vert_f_code;
	  stream->param.back_hor_f_code = stream->param.back_vert_f_code = 15;
	  sxf = stream->param.motion_data[0].sxf;
	  syf = stream->param.motion_data[0].syf;
	}

      motion_estimation (stream, stream->param.oldorgframe[0], stream->param.neworgframe[0],
		 stream->param.oldrefframe[0], stream->param.newrefframe[0],
			 neworg[0], newref[0],
		       sxf, syf, sxb, syb, stream->param.mbinfo, 1, ipflag);

      predict (stream,
	       stream->param.oldrefframe,
	       stream->param.newrefframe,
	       stream->param.predframe,
	       1,
	       stream->param.mbinfo);
      dct_type_estimation (stream, stream->param.predframe[0], neworg[0], stream->param.mbinfo);
      transform (stream, stream->param.predframe, neworg, stream->param.mbinfo, stream->param.blocks);

      stream->neworg = neworg[0];
      putpict (stream, neworg[0]);

      for (k = 0; k < stream->param.mb_height2 * stream->param.mb_width; k++)
	{
	  if (stream->param.mbinfo[k].mb_type & MB_INTRA)
	    for (j = 0; j < stream->param.block_count; j++)
	      iquant_intra (stream, stream->param.blocks[k * stream->param.block_count + j], stream->param.blocks[k * stream->param.block_count + j],
			    stream->param.dc_prec, stream->param.intra_q, stream->param.mbinfo[k].mquant);
	  else
	    for (j = 0; j < stream->param.block_count; j++)
	      iquant_non_intra (stream, stream->param.blocks[k * stream->param.block_count + j], stream->param.blocks[k * stream->param.block_count + j],
		     stream->param.inter_q, stream->param.mbinfo[k].mquant);
	}

      itransform (stream, stream->param.predframe, newref, stream->param.mbinfo, stream->param.blocks);
      calcSNR (stream, neworg, newref);
      stats (stream);
    }
  else
    {
      stream->param.pict_struct = FRAME_PICTURE;

      motion_estimation (stream, stream->param.oldorgframe[0], stream->param.neworgframe[0],
		 stream->param.oldrefframe[0], stream->param.newrefframe[0],
			 neworg[0], newref[0],
			 sxf, syf, sxb, syb, stream->param.mbinfo, 0, 0);




      predict (stream, stream->param.oldrefframe, stream->param.newrefframe, stream->param.predframe, 0, stream->param.mbinfo);


      dct_type_estimation (stream, stream->param.predframe[0], neworg[0], stream->param.mbinfo);


      transform (stream, stream->param.predframe, neworg, stream->param.mbinfo, stream->param.blocks);




      stream->neworg = neworg[0];
      /*  
         for (k=0; k<stream->param.mb_height*stream->param.mb_width; k++)
         {
         if (stream->param.mbinfo[k].mb_type & MB_INTRA)
         for (j=0; j<stream->param.block_count; j++)
         iquant_intra(stream,stream->param.blocks[k*stream->param.block_count+j],stream->param.blocks[k*stream->param.block_count+j],
         stream->param.dc_prec,stream->param.intra_q,stream->param.mbinfo[k].mquant);
         else
         for (j=0;j<stream->param.block_count;j++)
         iquant_non_intra(stream,stream->param.blocks[k*stream->param.block_count+j],stream->param.blocks[k*stream->param.block_count+j],
         stream->param.inter_q,stream->param.mbinfo[k].mquant);
         }

         itransform(stream,stream->param.predframe,newref,stream->param.mbinfo,stream->param.blocks);
         calcSNR(stream,neworg,newref);
       */
    }

  i++;
}
