/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include  "fdctref.h"
#include  "idct.h"
#include  "motion.h"
#include  "mpeg2enc.h"
#include  "params.h"
#include  "predict.h"
#include  "putbits.h"
#include  "puthdr.h"
#include  "putmpg.h"
#include  "putpic.h"
#include  "putseq.h"
#include  "putvlc.h"
#include  "quantize.h"
#include  "ratectl.h"
#include  "readpic.h"
#include  "stats.h"
#include  "transfrm.h"
#include  "writepic.h"
#include <stdio.h>
#include <stdlib.h>
#include "config.h"
#include "quant.h"

/* private prototypes */

static void border_extend _ANSI_ARGS_ ((unsigned char *frame, int w1, int h1,
					int w2, int h2));
static void conv444to422 (pm2v_stream * stream, unsigned char *src, unsigned char *dst);
static void conv422to420 (pm2v_stream * stream, unsigned char *src, unsigned char *dst);

void 
my_readframe (pm2v_stream * stream, unsigned char *frame[], unsigned char *img)
{
  
    
    int i, j, k;
  
    int r, g, b;
  
    double y, u, v;
  
    double cr, cg, cb, cu, cv;
  
    unsigned char *yp, *up, *vp;
  
    int mbc;
  static unsigned char *u444, *v444, *u422, *v422;
  
    static double coef[7][3] =
  {
   {0.2125, 0.7154, 0.0721}, /* ITU-R Rec. 709 (1990) */ 
   {0.299, 0.587, 0.114}, /* unspecified */ 
   {0.299, 0.587, 0.114}, /* reserved */ 
   {0.30, 0.59, 0.11}, /* FCC */ 
   {0.299, 0.587, 0.114}, /* ITU-R Rec. 624-4 System B, G */ 
   {0.299, 0.587, 0.114}, /* SMPTE 170M */ 
   {0.212, 0.701, 0.087}};	/* SMPTE 240M (1987) */
  
    



    pm2v_scatter_image (stream, img, stream->param.width * 3, stream->param.height);
  i = stream->param.matrix_coefficients;
  
    if (i > 8)
    
      i = 3;
  
    
    cr = coef[i - 1][0];
  
    cg = coef[i - 1][1];
  
    cb = coef[i - 1][2];
  
    cu = 0.5 / (1.0 - cb);
  
    cv = 0.5 / (1.0 - cr);
  
    
    if (stream->param.chroma_format == CHROMA444)
    
    {
      
	u444 = frame[1];
      
	v444 = frame[2];
      
    }
  
    else
    
    {
      
	if (!u444)
	
	{
	  
	    if (!(u444 = (unsigned char *) malloc (stream->param.width * stream->param.height)))
	    
	      error ("malloc failed");
	  
	    if (!(v444 = (unsigned char *) malloc (stream->param.width * stream->param.height)))
	    
	      error ("malloc failed");
	  
	    if (stream->param.chroma_format == CHROMA420)
	    
	    {
	      
		if (!(u422 = (unsigned char *) malloc ((stream->param.width >> 1) * stream->param.height)))
		
		  error ("malloc failed");
	      
		if (!(v422 = (unsigned char *) malloc ((stream->param.width >> 1) * stream->param.height)))
		
		  error ("malloc failed");
	      
	    }
	  
	}
      
    }
  
    
    
  /* skip header */ 
  /* magic number (P6) */ 
  /* width height maxcolors */ 


    

    for (i = 0, k = 0; i < stream->param.vertical_size; i++)
    
    {
      
	yp = frame[0] + i * stream->param.width;
      
	up = u444 + i * stream->param.width;
      
	vp = v444 + i * stream->param.width;
      
	
	if ((i >= (stream->offset_lines[stream->my_id] - 13))
	    && (i < (stream->offset_lines[stream->my_id + 1] + 13)))
	{
	  k = i * stream->param.horizontal_size * 3;
	  for (j = 0; j < stream->param.horizontal_size; j++, k += 3)
	    
	    {
	      
	      //fprintf(stderr,"%i \n",k);
	      //      mbc=(i/16)*stream->param.mb_width+j/16;

	      //  if (stream->my_id==stream->macroblock_to_process[mbc]) 
		r = img[k];
	      
		g = img[k + 1];
	      
		b = img[k + 2];
	      
	      /* convert to YUV */ 
		y = cr * r + cg * g + cb * b;
	      
		u = cu * (b - y);
	      
		v = cv * (r - y);
	      
		yp[j] = (219.0 / 256.0) * y + 16.5;	/* nominal range: 16..235 */
	      
		up[j] = (224.0 / 256.0) * u + 128.5;	/* 16..240 */
	      
		vp[j] = (224.0 / 256.0) * v + 128.5;	/* 16..240 */
	      
	    }
	}
      
    }
  
    
  //fprintf(stderr,"\n");
  //pm2v_gather_image(stream,u444,stream->param.width,stream->param.height);
  //pm2v_gather_image(stream,v444,stream->param.width,stream->param.height);

  //pm2v_exchange_lumi(stream,u444);
  //pm2v_exchange_lumi(stream,v444);
  //pm2v_exchange_lumi(stream,frame[0]);
    border_extend (frame[0], stream->param.horizontal_size, stream->param.vertical_size, stream->param.width, stream->param.height);
  
    border_extend (u444, stream->param.horizontal_size, stream->param.vertical_size, stream->param.width, stream->param.height);
  
    border_extend (v444, stream->param.horizontal_size, stream->param.vertical_size, stream->param.width, stream->param.height);
  
    
    pm2v_gather_image (stream, frame[0], stream->param.width, stream->param.height);
  if (stream->param.chroma_format == CHROMA422)
    
    {
      
	conv444to422 (stream, u444, frame[1]);
      
	conv444to422 (stream, v444, frame[2]);
      
	pm2v_gather_image (stream, u444, stream->param.width / 2, stream->param.height);
      pm2v_gather_image (stream, v444, stream->param.width / 2, stream->param.height);

      //pm2v_reduce_chroma(stream,frame[1],CHROMA422);
      //pm2v_reduce_chroma(stream,frame[2],CHROMA422);
    }
  
    
    if (stream->param.chroma_format == CHROMA420)
    
    {
      
	conv444to422 (stream, u444, u422);
      
	conv444to422 (stream, v444, v422);
      

	conv422to420 (stream, u422, frame[1]);
      
	conv422to420 (stream, v422, frame[2]);
      
	pm2v_gather_image (stream, frame[1]
			,stream->param.width / 2, stream->param.height / 2);
      pm2v_gather_image (stream, frame[2]
			 ,stream->param.width / 2, stream->param.height / 2);
    }
  
    
    

  //pm2v_reduce_lumi(stream,frame[0]);

}





static void 
border_extend (frame, w1, h1, w2, h2)
     unsigned char *frame;
     int w1, h1, w2, h2;
{
  int i, j;
  unsigned char *fp;

  /* horizontal pixel replication (right border) */

  for (j = 0; j < h1; j++)
    {
      fp = frame + j * w2;
      for (i = w1; i < w2; i++)
	fp[i] = fp[i - 1];
    }

  /* vertical pixel replication (bottom border) */

  for (j = h1; j < h2; j++)
    {
      fp = frame + j * w2;
      for (i = 0; i < w2; i++)
	fp[i] = fp[i - w2];
    }
}

/* horizontal filter and 2:1 subsampling */
static void 
conv444to422 (stream, src, dst)
     pm2v_stream *stream;
     unsigned char *src, *dst;
{
  int i, j, im5, im4, im3, im2, im1, ip1, ip2, ip3, ip4, ip5, ip6;


  if (stream->param.mpeg1)
    {
      for (j = 0; j < stream->param.height; j++)
	{
	  if ((j >= (stream->offset_lines[stream->my_id] - 13))
	      && (j < (stream->offset_lines[stream->my_id + 1] + 13)))
	    for (i = 0; i < stream->param.width; i += 2)
	      {
		im5 = (i < 5) ? 0 : i - 5;
		im4 = (i < 4) ? 0 : i - 4;
		im3 = (i < 3) ? 0 : i - 3;
		im2 = (i < 2) ? 0 : i - 2;
		im1 = (i < 1) ? 0 : i - 1;
		ip1 = (i < stream->param.width - 1) ? i + 1 : stream->param.width - 1;
		ip2 = (i < stream->param.width - 2) ? i + 2 : stream->param.width - 1;
		ip3 = (i < stream->param.width - 3) ? i + 3 : stream->param.width - 1;
		ip4 = (i < stream->param.width - 4) ? i + 4 : stream->param.width - 1;
		ip5 = (i < stream->param.width - 5) ? i + 5 : stream->param.width - 1;
		ip6 = (i < stream->param.width - 5) ? i + 6 : stream->param.width - 1;

		/* FIR filter with 0.5 sample interval phase shift */
		dst[i >> 1] = stream->param.clp[(int) (228 * (src[i] + src[ip1])
						+ 70 * (src[im1] + src[ip2])
						- 37 * (src[im2] + src[ip3])
						- 21 * (src[im3] + src[ip4])
						+ 11 * (src[im4] + src[ip5])
				   + 5 * (src[im5] + src[ip6]) + 256) >> 9];

	      }
	  src += stream->param.width;
	  dst += stream->param.width >> 1;
	}
    }
  else
    {
      /* MPEG-2 */
      for (j = 0; j < stream->param.height; j++)
	{
	  if ((j >= (stream->offset_lines[stream->my_id] - 13))
	      && (j < (stream->offset_lines[stream->my_id + 1] + 13)))
	    for (i = 0; i < stream->param.width; i += 2)
	      {
		im5 = (i < 5) ? 0 : i - 5;
		im3 = (i < 3) ? 0 : i - 3;
		im1 = (i < 1) ? 0 : i - 1;
		ip1 = (i < stream->param.width - 1) ? i + 1 : stream->param.width - 1;
		ip3 = (i < stream->param.width - 3) ? i + 3 : stream->param.width - 1;
		ip5 = (i < stream->param.width - 5) ? i + 5 : stream->param.width - 1;

		/* FIR filter coefficients (*512): 22 0 -52 0 159 256 159 0 -52 0 22 */
		dst[i >> 1] = stream->param.clp[(int) (22 * (src[im5] + src[ip5]) - 52 * (src[im3] + src[ip3])
		  + 159 * (src[im1] + src[ip1]) + 256 * src[i] + 256) >> 9];

	      }
	  src += stream->param.width;
	  dst += stream->param.width >> 1;
	}
    }
}

/* vertical filter and 2:1 subsampling */
static void 
conv422to420 (stream, src, dst)
     pm2v_stream *stream;
     unsigned char *src, *dst;
{
  int w, i, j, jm6, jm5, jm4, jm3, jm2, jm1;
  int jp1, jp2, jp3, jp4, jp5, jp6;


  w = stream->param.width >> 1;

  if (stream->param.prog_frame)
    {
      /* intra frame */
      for (i = 0; i < w; i++)
	{
	  for (j = 0; j < stream->param.height; j += 2)
	    if ((j >= stream->offset_lines[stream->my_id])
		&& (j < stream->offset_lines[stream->my_id + 1]))
	      {

		jm5 = (j < 5) ? 0 : j - 5;
		jm4 = (j < 4) ? 0 : j - 4;
		jm3 = (j < 3) ? 0 : j - 3;
		jm2 = (j < 2) ? 0 : j - 2;
		jm1 = (j < 1) ? 0 : j - 1;
		jp1 = (j < stream->param.height - 1) ? j + 1 : stream->param.height - 1;
		jp2 = (j < stream->param.height - 2) ? j + 2 : stream->param.height - 1;
		jp3 = (j < stream->param.height - 3) ? j + 3 : stream->param.height - 1;
		jp4 = (j < stream->param.height - 4) ? j + 4 : stream->param.height - 1;
		jp5 = (j < stream->param.height - 5) ? j + 5 : stream->param.height - 1;
		jp6 = (j < stream->param.height - 5) ? j + 6 : stream->param.height - 1;

		/* FIR filter with 0.5 sample interval phase shift */
		dst[w * (j >> 1)] = stream->param.clp[(int) (228 * (src[w * j] + src[w * jp1])
					+ 70 * (src[w * jm1] + src[w * jp2])
					- 37 * (src[w * jm2] + src[w * jp3])
					- 21 * (src[w * jm3] + src[w * jp4])
					+ 11 * (src[w * jm4] + src[w * jp5])
			   + 5 * (src[w * jm5] + src[w * jp6]) + 256) >> 9];

	      }
	  src++;
	  dst++;
	}
    }
  else
    {
      /* intra field */
      for (i = 0; i < w; i++)
	{

	  for (j = 0; j < stream->param.height; j += 4)
	    if ((j >= stream->offset_lines[stream->my_id])
		&& (j < stream->offset_lines[stream->my_id + 1]))
	      {
		/* top field */
		jm5 = (j < 10) ? 0 : j - 10;
		jm4 = (j < 8) ? 0 : j - 8;
		jm3 = (j < 6) ? 0 : j - 6;
		jm2 = (j < 4) ? 0 : j - 4;
		jm1 = (j < 2) ? 0 : j - 2;
		jp1 = (j < stream->param.height - 2) ? j + 2 : stream->param.height - 2;
		jp2 = (j < stream->param.height - 4) ? j + 4 : stream->param.height - 2;
		jp3 = (j < stream->param.height - 6) ? j + 6 : stream->param.height - 2;
		jp4 = (j < stream->param.height - 8) ? j + 8 : stream->param.height - 2;
		jp5 = (j < stream->param.height - 10) ? j + 10 : stream->param.height - 2;
		jp6 = (j < stream->param.height - 12) ? j + 12 : stream->param.height - 2;

		/* FIR filter with 0.25 sample interval phase shift */
		dst[w * (j >> 1)] = stream->param.clp[(int) (8 * src[w * jm5]
							  + 5 * src[w * jm4]
							 - 30 * src[w * jm3]
							 - 18 * src[w * jm2]
							+ 113 * src[w * jm1]
							  + 242 * src[w * j]
							+ 192 * src[w * jp1]
							 + 35 * src[w * jp2]
							 - 38 * src[w * jp3]
							 - 10 * src[w * jp4]
							 + 11 * src[w * jp5]
					    + 2 * src[w * jp6] + 256) >> 9];

		/* bottom field */
		jm6 = (j < 9) ? 1 : j - 9;
		jm5 = (j < 7) ? 1 : j - 7;
		jm4 = (j < 5) ? 1 : j - 5;
		jm3 = (j < 3) ? 1 : j - 3;
		jm2 = (j < 1) ? 1 : j - 1;
		jm1 = (j < stream->param.height - 1) ? j + 1 : stream->param.height - 1;
		jp1 = (j < stream->param.height - 3) ? j + 3 : stream->param.height - 1;
		jp2 = (j < stream->param.height - 5) ? j + 5 : stream->param.height - 1;
		jp3 = (j < stream->param.height - 7) ? j + 7 : stream->param.height - 1;
		jp4 = (j < stream->param.height - 9) ? j + 9 : stream->param.height - 1;
		jp5 = (j < stream->param.height - 11) ? j + 11 : stream->param.height - 1;
		jp6 = (j < stream->param.height - 13) ? j + 13 : stream->param.height - 1;

		/* FIR filter with 0.25 sample interval phase shift */
		dst[w * ((j >> 1) + 1)] = stream->param.clp[(int) (8 * src[w * jp6]
							  + 5 * src[w * jp5]
							 - 30 * src[w * jp4]
							 - 18 * src[w * jp3]
							+ 113 * src[w * jp2]
							+ 242 * src[w * jp1]
							+ 192 * src[w * jm1]
							 + 35 * src[w * jm2]
							 - 38 * src[w * jm3]
							 - 10 * src[w * jm4]
							 + 11 * src[w * jm5]
					    + 2 * src[w * jm6] + 256) >> 9];

	      }
	  src++;
	  dst++;
	}
    }
}

/* pbm_getc() and pbm_getint() are essentially taken from
 * PBMPLUS (C) Jef Poskanzer
 * but without error/EOF checking
 */
char 
pbm_getc (file)
     FILE *file;
{
  char ch;

  ch = getc (file);

  if (ch == '#')
    {
      do
	{
	  ch = getc (file);
	}
      while (ch != '\n' && ch != '\r');
    }

  return ch;
}

int 
pbm_getint (file)
     FILE *file;
{
  char ch;
  int i;

  do
    {
      ch = pbm_getc (file);
    }
  while (ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r');

  i = 0;
  do
    {
      i = i * 10 + ch - '0';
      ch = pbm_getc (file);
    }
  while (ch >= '0' && ch <= '9');

  return i;
}
