/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include  "fdctref.h"
#include  "idct.h"
#include  "motion.h"
#include  "mpeg2enc.h"
#include  "params.h"
#include  "predict.h"
#include  "putbits.h"
#include  "puthdr.h"
#include  "putmpg.h"
#include  "putpic.h"
#include  "putseq.h"
#include  "putvlc.h"
#include  "quantize.h"
#include  "ratectl.h"
#include  "readpic.h"
#include  "stats.h"
#include  "transfrm.h"
#include  "writepic.h"
#include <stdio.h>
#include <math.h>
#include "config.h"
#include "quant.h"

/* private prototypes */
static void calcSNR1 (pm2v_stream * stream, unsigned char *org, unsigned char *rec,
		      int lx, int w, int h, double *pv, double *pe);


void 
calcSNR (pm2v_stream * stream, unsigned char *org[3], unsigned char *rec[3])
{
  int w, h, offs;
  double v, e;

  w = stream->param.horizontal_size;
  h = (stream->param.pict_struct == FRAME_PICTURE) ? stream->param.vertical_size : (stream->param.vertical_size >> 1);
  offs = (stream->param.pict_struct == BOTTOM_FIELD) ? stream->param.width : 0;

  calcSNR1 (stream, org[0] + offs, rec[0] + offs, stream->param.width2, w, h, &v, &e);


  if (stream->param.chroma_format != CHROMA444)
    {
      w >>= 1;
      offs >>= 1;
    }

  if (stream->param.chroma_format == CHROMA420)
    h >>= 1;

  calcSNR1 (stream, org[1] + offs, rec[1] + offs, stream->param.chrom_width2, w, h, &v, &e);


  calcSNR1 (stream, org[2] + offs, rec[2] + offs, stream->param.chrom_width2, w, h, &v, &e);

}

static void 
calcSNR1 (stream, org, rec, lx, w, h, pv, pe)
     pm2v_stream *stream;
     unsigned char *org;
     unsigned char *rec;
     int lx, w, h;
     double *pv, *pe;
{
  int i, j;
  double v1, s1, s2, e2;

  s1 = s2 = e2 = 0.0;

  for (j = 0; j < h; j++)
    {
      for (i = 0; i < w; i++)
	{
	  v1 = org[i];
	  s1 += v1;
	  s2 += v1 * v1;
	  v1 -= rec[i];
	  e2 += v1 * v1;
	}
      org += lx;
      rec += lx;
    }

  s1 /= w * h;
  s2 /= w * h;
  e2 /= w * h;

  /* prevent division by zero in calcSNR() */
  if (e2 == 0.0)
    e2 = 0.00001;

  *pv = s2 - s1 * s1;		/* variance */
  *pe = e2;			/* MSE */
}

void 
stats (pm2v_stream * stream)
{
  int i, k, nmb;
  int n_skipped, n_intra, n_ncoded, n_blocks, n_interp, n_forward, n_backward;
  struct mbinfo *mbi;

  nmb = stream->param.mb_width * stream->param.mb_height2;

  n_skipped = n_intra = n_ncoded = n_blocks = n_interp = n_forward = n_backward = 0;

  for (k = 0; k < nmb; k++)
    {
      mbi = stream->param.mbinfo + k;
      if (mbi->skipped)
	n_skipped++;
      else if (mbi->mb_type & MB_INTRA)
	n_intra++;
      else if (!(mbi->mb_type & MB_PATTERN))
	n_ncoded++;

      for (i = 0; i < stream->param.block_count; i++)
	if (mbi->cbp & (1 << i))
	  n_blocks++;

      if (mbi->mb_type & MB_FORWARD)
	{
	  if (mbi->mb_type & MB_BACKWARD)
	    n_interp++;
	  else
	    n_forward++;
	}
      else if (mbi->mb_type & MB_BACKWARD)
	n_backward++;
    }

  /*

     k = 0;

     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     mbi = stream->param.mbinfo + k;
     mb_type = mbi->mb_type;
     if (mbi->skipped)
     putc('S',statfile);
     else if (mb_type & MB_INTRA)
     putc('I',statfile);
     else switch (mb_type & (MB_FORWARD|MB_BACKWARD))
     {
     case MB_FORWARD:
     putc(mbi->motion_type==MC_FIELD ? 'f' :
     mbi->motion_type==MC_DMV   ? 'p' :
     'F',statfile); break;
     case MB_BACKWARD:
     putc(mbi->motion_type==MC_FIELD ? 'b' :
     'B',statfile); break;
     case MB_FORWARD|MB_BACKWARD:
     putc(mbi->motion_type==MC_FIELD ? 'd' :
     'D',statfile); break;
     default:
     putc('0',statfile); break;
     }

     if (mb_type & MB_QUANT)
     putc('Q',statfile);
     else if (mb_type & (MB_PATTERN|MB_INTRA))
     putc(' ',statfile);
     else
     putc('N',statfile);

     putc(' ',statfile);

     k++;
     }
     putc('\n',statfile);
     }

     fprintf(statfile,"\nmquant map:\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (i==0 || stream->param.mbinfo[k].mquant!=stream->param.mbinfo[k-1].mquant)
     fprintf(statfile,"%3d",stream->param.mbinfo[k].mquant);
     else
     fprintf(statfile,"   ");

     k++;
     }
     putc('\n',statfile);
     }

     #if 0
     fprintf(statfile,"\ncbp map:\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     fprintf(statfile,"%02x ",stream->param.mbinfo[k].cbp);

     k++;
     }
     putc('\n',statfile);
     }

     if (stream->param.pict_struct==FRAME_PICTURE && !frame_pred_dct)
     {
     fprintf(statfile,"\ndct_type map:\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & (MB_PATTERN|MB_INTRA))
     fprintf(statfile,"%d  ",stream->param.mbinfo[k].dct_type);
     else
     fprintf(statfile,"   ");

     k++;
     }
     putc('\n',statfile);
     }
     }

     if (pict_type!=I_TYPE)
     {
     fprintf(statfile,"\nforward motion vectors (first vector, horizontal):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_FORWARD)
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[0][0][0]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }

     fprintf(statfile,"\nforward motion vectors (first vector, vertical):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_FORWARD)
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[0][0][1]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }

     fprintf(statfile,"\nforward motion vectors (second vector, horizontal):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_FORWARD
     && ((stream->param.pict_struct==FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_FIELD) ||
     (stream->param.pict_struct!=FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_16X8)))
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[1][0][0]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }

     fprintf(statfile,"\nforward motion vectors (second vector, vertical):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_FORWARD
     && ((stream->param.pict_struct==FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_FIELD) ||
     (stream->param.pict_struct!=FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_16X8)))
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[1][0][1]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }


     }

     if (pict_type==B_TYPE)
     {
     fprintf(statfile,"\nbackward motion vectors (first vector, horizontal):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_BACKWARD)
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[0][1][0]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }

     fprintf(statfile,"\nbackward motion vectors (first vector, vertical):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_BACKWARD)
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[0][1][1]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }

     fprintf(statfile,"\nbackward motion vectors (second vector, horizontal):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_BACKWARD
     && ((stream->param.pict_struct==FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_FIELD) ||
     (stream->param.pict_struct!=FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_16X8)))
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[1][1][0]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }

     fprintf(statfile,"\nbackward motion vectors (second vector, vertical):\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     if (stream->param.mbinfo[k].mb_type & MB_BACKWARD
     && ((stream->param.pict_struct==FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_FIELD) ||
     (stream->param.pict_struct!=FRAME_PICTURE && stream->param.mbinfo[k].motion_type==MC_16X8)))
     fprintf(statfile,"%4d",stream->param.mbinfo[k].MV[1][1][1]);
     else
     fprintf(statfile,"   .");

     k++;
     }
     putc('\n',statfile);
     }


     }
     #endif

     #if 0
     //useful for debugging 
     fprintf(statfile,"\nmacroblock info dump:\n");

     k=0;
     for (j=0; j<stream->param.mb_height2; j++)
     {
     for (i=0; i<stream->param.mb_width; i++)
     {
     fprintf(statfile,"%d: %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
     k,
     stream->param.mbinfo[k].mb_type,
     stream->param.mbinfo[k].motion_type,
     stream->param.mbinfo[k].dct_type,
     stream->param.mbinfo[k].mquant,
     stream->param.mbinfo[k].cbp,
     stream->param.mbinfo[k].skipped,
     stream->param.mbinfo[k].MV[0][0][0],
     stream->param.mbinfo[k].MV[0][0][1],
     stream->param.mbinfo[k].MV[0][1][0],
     stream->param.mbinfo[k].MV[0][1][1],
     stream->param.mbinfo[k].MV[1][0][0],
     stream->param.mbinfo[k].MV[1][0][1],
     stream->param.mbinfo[k].MV[1][1][0],
     stream->param.mbinfo[k].MV[1][1][1],
     stream->param.mbinfo[k].mv_field_sel[0][0],
     stream->param.mbinfo[k].mv_field_sel[0][1],
     stream->param.mbinfo[k].mv_field_sel[1][0],
     stream->param.mbinfo[k].mv_field_sel[1][1]);

     k++;
     }
     }
     #endif
   */
}
