/*****************************************************************************
Original version of mpeg2encode is copyrighted to:
MPEG Software Simulation Group. All Rights Reserved.

Disclaimer of Warranty

mpeg2encode is available to the user without any license fee or
royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
any and all warranties, whether express, implied, or statuary, including any
implied warranties or merchantability or of fitness for a particular
purpose.  In no event shall the copyright-holder be liable for any
incidental, punitive, or consequential damages of any kind whatsoever
arising from the use of these programs.

This disclaimer of warranty extends to the user of these programs and user's
customers, employees, agents, transferees, successors, and assigns.

The MPEG Software Simulation Group does not represent or warrant that the
programs furnished hereunder are free of infringement of any third-party
patents.

Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
are subject to royalty fees to patent holders.  Many of these patents are
general enough such that they are unavoidable regardless of implementation
design.


Modification of mpeg2encode contained in PM2V source code is copyrighted by 
Jean Laurent and Olivier Guillot.
This modification are distributed under the GPL (GNU Public License). 
see license.txt in the doc directory for details.

You can contact us to pm2v@free.fr or pm2v@sourceforge.net
*****************************************************************************/

#include  "fdctref.h" 
#include  "idct.h" 
#include  "motion.h" 
#include  "mpeg2enc.h" 
#include  "params.h" 
#include  "predict.h" 
#include  "putbits.h" 
#include  "puthdr.h" 
#include  "putmpg.h" 
#include  "putpic.h" 
#include  "putseq.h" 
#include  "putvlc.h" 
#include  "quantize.h" 
#include  "ratectl.h" 
#include  "readpic.h" 
#include  "stats.h" 
#include  "transfrm.h"
#include  "writepic.h" 
#include <stdio.h>
#include <math.h>
#include "config.h"
#include "quant.h"

/* private prototypes*/
static void add_pred (pm2v_stream* stream,unsigned char *pred, unsigned char *cur,
  int lx, short *blk);
static void sub_pred (unsigned char *pred, unsigned char *cur,
  int lx, short *blk);

/* subtract prediction and transform prediction error */
void transform(pm2v_stream* stream,unsigned char *pred[],unsigned char *cur[],struct mbinfo *mbi,short blocks[][64])
{
  int i, j, i1, j1, k, n, cc, offs, lx;

  k = 0;
  for (j=0; j<stream->param.height2; j+=16)
    for (i=0; i<stream->param.width; i+=16)
    {
      if (stream->my_id==stream->macroblock_to_process[k])
	{
	  //for (n=(stream->stride_blocks[stream->my_id]/(sizeof(short)*64)); 
	  //  n<(stream->stride_blocks[stream->my_id+1]/(sizeof(short)*64)); n++)
	  for (n=0; n<stream->param.block_count; n++)
	    {
	      cc = (n<4) ? 0 : (n&1)+1; /* color component index */
	      if (cc==0)
		{
		  /* luminance */
		  if ((stream->param.pict_struct==FRAME_PICTURE) && mbi[k].dct_type)
		    {
		      /* field DCT */
		      offs = i + ((n&1)<<3) + stream->param.width*(j+((n&2)>>1));
		      lx = stream->param.width<<1;
		    }
		  else
		    {
		      /* frame DCT */
		      offs = i + ((n&1)<<3) + stream->param.width2*(j+((n&2)<<2));
		      lx = stream->param.width2;
		    }
		  
		  if (stream->param.pict_struct==BOTTOM_FIELD)
		    offs += stream->param.width;
		}
	      else
		{
		  /* chrominance */
		  
		  /* scale coordinates */
		  i1 = (stream->param.chroma_format==CHROMA444) ? i : i>>1;
		  j1 = (stream->param.chroma_format!=CHROMA420) ? j : j>>1;
		  
		  if ((stream->param.pict_struct==FRAME_PICTURE) && mbi[k].dct_type
		      && (stream->param.chroma_format!=CHROMA420))
		    {
		      /* field DCT */
		      offs = i1 + (n&8) + stream->param.chrom_width*(j1+((n&2)>>1));
		      lx = stream->param.chrom_width<<1;
		    }
		  else
		    {
		      /* frame DCT */
		      offs = i1 + (n&8) + stream->param.chrom_width2*(j1+((n&2)<<2));
		      lx = stream->param.chrom_width2;
		    }
		  
		  if (stream->param.pict_struct==BOTTOM_FIELD)
		    offs += stream->param.chrom_width;
		}
	      //if ((k>=stream->stride[stream->my_id]) && (k<stream->stride[stream->my_id+1]))
	      sub_pred(pred[cc]+offs,cur[cc]+offs,lx,blocks[k*stream->param.block_count+n]);
	      //if ((k>=stream->stride[stream->my_id]) && (k<stream->stride[stream->my_id+1]))
	      fdct(blocks[k*stream->param.block_count+n]);
	    }
	}
      k++;
	
    }
  pm2v_reduce_block(stream);
}

/* inverse transform prediction error and add prediction */
void itransform(pm2v_stream* stream,unsigned char *pred[],unsigned char *cur[],struct mbinfo *mbi,short blocks[][64])
{
  int i, j, i1, j1, k, n, cc, offs, lx;

  k = 0;

  for (j=0; j<stream->param.height2; j+=16)
    for (i=0; i<stream->param.width; i+=16)
    {
      for (n=0; n<stream->param.block_count; n++)
      {
        cc = (n<4) ? 0 : (n&1)+1; /* color component index */

        if (cc==0)
        {
          /* luminance */
          if ((stream->param.pict_struct==FRAME_PICTURE) && mbi[k].dct_type)
          {
            /* field DCT */
            offs = i + ((n&1)<<3) + stream->param.width*(j+((n&2)>>1));
            lx = stream->param.width<<1;
          }
          else
          {
            /* frame DCT */
            offs = i + ((n&1)<<3) + stream->param.width2*(j+((n&2)<<2));
            lx = stream->param.width2;
          }

          if (stream->param.pict_struct==BOTTOM_FIELD)
            offs += stream->param.width;
        }
        else
        {
          /* chrominance */

          /* scale coordinates */
          i1 = (stream->param.chroma_format==CHROMA444) ? i : i>>1;
          j1 = (stream->param.chroma_format!=CHROMA420) ? j : j>>1;

          if ((stream->param.pict_struct==FRAME_PICTURE) && mbi[k].dct_type
              && (stream->param.chroma_format!=CHROMA420))
          {
            /* field DCT */
            offs = i1 + (n&8) + stream->param.chrom_width*(j1+((n&2)>>1));
            lx = stream->param.chrom_width<<1;
          }
          else
          {
            /* frame DCT */
            offs = i1 + (n&8) + stream->param.chrom_width2*(j1+((n&2)<<2));
            lx = stream->param.chrom_width2;
          }

          if (stream->param.pict_struct==BOTTOM_FIELD)
            offs += stream->param.chrom_width;
        }

        idct(blocks[k*stream->param.block_count+n]);
        add_pred(stream,pred[cc]+offs,cur[cc]+offs,lx,blocks[k*stream->param.block_count+n]);
      }

      k++;
    }
}

/* add prediction and prediction error, saturate to 0...255 */
static void add_pred(pm2v_stream* stream,unsigned char *pred,unsigned char *cur,int lx,short *blk)
{
  int i, j;

  for (j=0; j<8; j++)
  {
    for (i=0; i<8; i++)
      cur[i] = stream->param.clp[blk[i] + pred[i]];
    blk+= 8;
    cur+= lx;
    pred+= lx;
  }
}

/* subtract prediction from block data */
static void sub_pred(unsigned char *pred,unsigned char *cur,int lx,short *blk)
{
  int i, j;

  for (j=0; j<8; j++)
  {
    for (i=0; i<8; i++)
      blk[i] = cur[i] - pred[i];
    blk+= 8;
    cur+= lx;
    pred+= lx;
  }
}

/*
 * select between frame and field DCT
 *
 * preliminary version: based on inter-field correlation
 */
void dct_type_estimation(pm2v_stream *stream,unsigned char *pred,unsigned char*cur,struct mbinfo *mbi)
{
  short blk0[128], blk1[128];
  int i, j, i0, j0, k, offs, s0, s1, sq0, sq1, s01;
  double d, r;

  k = 0;

  for (j0=0; j0<stream->param.height2; j0+=16)
    for (i0=0; i0<stream->param.width; i0+=16)
    {
      if (stream->my_id==stream->macroblock_to_process[k])
	{
	  if (stream->param.frame_pred_dct || stream->param.pict_struct!=FRAME_PICTURE)
	    mbi[k].dct_type = 0;
	  else
	    {
	      /* interlaced frame picture */
	      /*
	       * calculate prediction error (cur-pred) for top (blk0)
	       * and bottom field (blk1)
	       */
	      for (j=0; j<8; j++)
		{
		  offs = stream->param.width*((j<<1)+j0) + i0;
		  for (i=0; i<16; i++)
		    {
		      blk0[16*j+i] = cur[offs] - pred[offs];
		      blk1[16*j+i] = cur[offs+stream->param.width] - pred[offs+stream->param.width];
		      offs++;
		    }
		}
	      /* correlate fields */
	      s0=s1=sq0=sq1=s01=0;
	      
	      for (i=0; i<128; i++)
		{
		  s0+= blk0[i];
		  sq0+= blk0[i]*blk0[i];
		  s1+= blk1[i];
		  sq1+= blk1[i]*blk1[i];
		  s01+= blk0[i]*blk1[i];
		}
	      
	      d = (sq0-(s0*s0)/128.0)*(sq1-(s1*s1)/128.0);
	      
	      if (d>0.0)
		{
		  r = (s01-(s0*s1)/128.0)/sqrt(d);
		  if (r>0.5)
		    mbi[k].dct_type = 0; /* frame DCT */
		  else
		    mbi[k].dct_type = 1; /* field DCT */
		}
	      else
		mbi[k].dct_type = 1; /* field DCT */
	    }
	  k++;
	}
    }
  pm2v_reduce_macroblock(stream);
}











